# -*- coding: utf-8 -*-
# TODO implement continuum and radial velocity tests

from itertools import product

import numpy as np
import pytest

from pysme.continuum_and_radial_velocity import (
    apply_radial_velocity_and_continuum,
    match_rv_continuum,
)


def test_match_both(testcase1):
    sme, x_syn, y_syn, rv = testcase1

    # Fix random results of the MCMC
    np.random.seed(0)

    vrad_options = ["none", "fix", "each", "whole"]
    cscale_options = ["none", "fix", "constant", "linear"]  # quadratic
    cscale_types = [
        "mask",
        "match",
        "match+mask",
        "mcmc",
        "spline",
        "spline+mask",
    ]

    # vrad_options = ["whole"]
    # cscale_options = ["linear"]  # quadratic
    # cscale_types = ["spline", "spline+mask"]
    segments = [[0], range(sme.nseg)]

    for voption, coption, ctype, segment in product(
        vrad_options, cscale_options, cscale_types, segments
    ):
        sme.vrad_flag = voption
        sme.cscale_flag = coption
        sme.cscale_type = ctype
        sme.vrad = None
        sme.cscale = None

        # Determine the radial velocity and the continuum
        cscale, cunc, vrad, vunc = match_rv_continuum(
            sme, segment, x_syn[segment], y_syn[segment]
        )

        # check that it can be applied
        smod = apply_radial_velocity_and_continuum(
            sme.wave, x_syn, y_syn, vrad, cscale, ctype, segment, copy=True
        )

        assert vrad is not None
        assert vunc is not None
        assert cscale is not None
        assert cunc is not None

        assert vrad.ndim == 1
        assert vrad.shape[0] == sme.nseg
        assert vunc.ndim == 2
        assert vunc.shape[0] == sme.nseg
        assert vunc.shape[1] == 2

        assert cscale.ndim == 2
        assert cscale.shape[0] == sme.nseg
        if ctype in ["mask", "match", "match+mask", "mcmc"]:
            assert cscale.shape[1] == sme.cscale_degree + 1
        elif ctype in ["spline", "spline+mask"]:
            assert np.allclose(cscale.shape[1], sme.cscale_degree)
        else:
            pass

        if ctype in ["mask", "match", "match+mask", "mcmc"]:
            assert cunc.ndim == 3
            assert cunc.shape[0] == sme.nseg
            assert cunc.shape[1] == sme.cscale_degree + 1
            assert cunc.shape[2] == 2
        elif ctype in ["spline", "spline+mask"]:
            assert cunc.ndim == 2
            assert cunc.shape[0] == sme.nseg
            assert np.allclose(cunc.shape[1], sme.cscale_degree)
        else:
            pass

        if voption in ["none", "fix"]:
            assert np.all(vrad == 0)
        else:
            for seg in range(sme.nseg):
                if seg in segment:
                    assert np.allclose(vrad[seg], rv, atol=1)
                else:
                    assert vrad[seg] == 0

        if ctype in ["mask", "match", "match+mask", "mcmc"]:
            if coption in ["none", "fix"]:
                assert np.all(cscale[:, -1] == 1)
                assert np.all(cscale[:, :-1] == 0)
            else:
                assert np.allclose(cscale[:, -1], 1, atol=1e-1)
                assert np.allclose(cscale[:, :-1], 0, atol=1e-1)
        elif ctype in ["spline", "spline+mask"]:
            if coption in ["none", "fix"]:
                assert np.all(cscale == 1)
            else:
                pass
        else:
            pass


# def test_nomask(testcase1):
#     sme, x_syn, y_syn, rv = testcase1
#     sme.cscale_flag = "constant"
#     sme.vrad_flag = "each"
#     sme.cscale_type = "mask"

#     sme.mask = 0
#     with pytest.warns(UserWarning):
#         rvel, vunc, cscale, cunc = match_rv_continuum(sme, 0, x_syn, y_syn)



# -*- coding: utf-8 -*-
import logging
from copy import copy
from multiprocessing.sharedctypes import Value

import numpy as np
import pybtex
from flex.extensions.bindata import MultipleDataExtension
from flex.flex import FlexExtension

from . import __file_ending__, __version__, persistence
from .iliffe_vector import Iliffe_vector

logger = logging.getLogger(__name__)


def this(self, x):
    """This just returns the input"""
    return x


def notNone(func):
    def f(self, value):
        return func(self, value) if value is not None else None

    return f


def uppercase(func):
    def f(self, value):
        try:
            value = value.upper()
        except AttributeError:
            pass
        return func(self, value)

    return f


def lowercase(func):
    def f(self, value):
        try:
            value = value.casefold()
        except AttributeError:
            pass
        return func(self, value)

    return f


def oneof(*options, astype=None):
    def f(self, value):
        if value not in options:
            if astype is not None and astype(value) == value:
                return astype(value)
            raise ValueError(f"Received {value} but expected one of {options}")
        return value

    return f


def astype(func, allow_None=False):
    def f(self, value):
        if allow_None and value is None:
            return value
        if isinstance(value, func):
            return value
        return func(value)

    return f


asint = astype(int)
asfloat = astype(float)
asstr = astype(str)
asbool = astype(bool)


def absolute(self, value):
    try:
        return abs(value)
    except TypeError:
        logger.warning(
            "Tried setting number field with %s, trying to use the 0th element of it instead",
            type(value),
        )
        return abs(value[0])


def array(shape, dtype, allow_None=True):
    special_shape = False
    if isinstance(shape, str):
        shape = shape.split(",")
        for i, s in enumerate(shape):
            try:
                shape[i] = int(s)
            except ValueError:
                special_shape = True

    def f(self, value):
        if value is not None:
            value = np.asarray(value, dtype=dtype)
            value = np.atleast_1d(value)

            if shape is not None:
                if special_shape:
                    # Replace string placeholders with actual values
                    newshape = [
                        getattr(self, s) if isinstance(s, str) else s for s in shape
                    ]
                else:
                    newshape = shape
                value = value.reshape(newshape)
        elif not allow_None:
            raise ValueError(f"Received None but expected datatype {dtype}")
        return value

    return f


def vector(self, value):
    if value is None:
        return None
    elif np.isscalar(value):
        if self.wave is None:
            raise ValueError(
                "Can not set vector from single value without defining the wavelength grid"
            )
        wind = self.wave.sizes
        value = np.full(np.sum(wind), value)
        value = Iliffe_vector.from_indices(value, wind)
    elif isinstance(value, np.ndarray):
        value = np.require(value, requirements="W")
        if value.ndim == 1:
            wind = self.wave.sizes if self.wave is not None else None
            value = Iliffe_vector.from_indices(value, wind)
        else:
            value = Iliffe_vector(value)
    elif isinstance(value, list):
        value = Iliffe_vector(value)
    elif isinstance(value, Iliffe_vector):
        pass
    elif isinstance(value, FlexExtension):
        value = Iliffe_vector._load(value)
    elif isinstance(value, np.lib.npyio.NpzFile):
        value = Iliffe_vector._load_v1(value)
    else:
        raise TypeError("Input value is of the wrong type")

    return value


# Shorter versions, but less obvious (also "_name" is calculated each time, instead of once?)
# fget = lambda name, func: lambda self: func(getattr(self, f"_{name}"))
# fset = lambda name, func: lambda self, value: setattr(self, f"_{name}", func(self, value))


def fget(name, func):
    name = f"_{name}"

    def f(self):
        return func(self, getattr(self, name))

    return f


def fset(name, func):
    name = f"_{name}"

    def f(self, value):
        setattr(self, name, func(self, value))

    return f


def CollectionFactory(cls):
    """Decorator that turns Collection _fields into properties"""

    # Add properties to the class
    for name, _, setter, getter, doc in cls._fields:
        setattr(
            cls,
            name,
            property(fget(name, getter), fset(name, setter), None, doc),
        )
    cls._names = [f[0] for f in cls._fields]

    return cls


@CollectionFactory
class Collection(persistence.IPersist):
    _fields = [
        (
            "citation_info",
            "",
            asstr,
            this,
            "str: Bibtex representation of the citation",
        )
    ]  # [("name", "default", str, this, "doc")]

    def __init__(self, **kwargs):
        for name, default, *_ in self._fields:
            setattr(self, name, copy(default))

        for key, value in kwargs.items():
            if key in self._names:
                if isinstance(value, bytes):
                    value = value.decode()
                elif isinstance(value, np.ndarray) and value.dtype == np.dtype("O"):
                    value = value.astype(str)
                setattr(self, key, value)

    def __getitem__(self, key):
        key = key.casefold()
        return getattr(self, key)

    def __setitem__(self, key, value):
        key = key.casefold()
        setattr(self, key, value)

    def __contains__(self, key):
        return key in dir(self) and getattr(self, key) is not None

    def __str__(self):
        with np.printoptions(threshold=10):
            lines = []
            for name in self._names:
                lines += [f"{name}={self[name]!r}"]
            lines = ",\n".join(lines)
            output = f"{self.__class__.__name__}({lines})"
        return output

    def __repr__(self):
        return self.__str__()

    def create_citation(self, citation_info, output="string"):
        if output == "bibtex":
            return citation_info
        elif output == "string":
            output = "plaintext"

        return pybtex.format_from_string(
            citation_info, style="plain", output_backend=output
        )

    def citation(self, output="string"):
        return self.create_citation(self.citation_info, output=output)

    def to_dict(self):
        tree = {}
        for name in self._names:
            value = self[name]
            if hasattr(value, "to_dict"):
                value = value.to_dict()
            tree[name] = value
        return tree

    @classmethod
    def from_dict(cls, data):
        obj = cls(**data)
        return obj

    def _save(self):
        header = {}
        data = {}

        for name in self._names:
            value = self[name]
            if isinstance(value, np.ndarray) and value.size > 10:
                data[name] = value
            else:
                header[name] = value

        ext = MultipleDataExtension(header, data)
        return ext

    @classmethod
    def _load(cls, ext: MultipleDataExtension):
        header = ext.header
        if isinstance(ext, MultipleDataExtension):
            data = dict(ext.data)
            header.update(data)

        obj = cls(**header)
        return obj



# -*- coding: utf-8 -*-
"""
Spectral Synthesis Module of SME
"""
import logging
import uuid

import numpy as np
from scipy.constants import speed_of_light
from scipy.interpolate import interp1d
from scipy.ndimage.filters import convolve
from tqdm import tqdm

from . import broadening
from .atmosphere.interpolation import AtmosphereInterpolator
from .continuum_and_radial_velocity import (
    apply_radial_velocity_and_continuum,
    match_rv_continuum,
    null_result,
)
from .iliffe_vector import Iliffe_vector
from .large_file_storage import setup_lfs
from .sme import MASK_VALUES
from .sme_synth import SME_DLL
from .util import show_progress_bars

logger = logging.getLogger(__name__)

clight = speed_of_light * 1e-3  # km/s

__DLL_DICT__ = {}
__DLL_IDS__ = {}


class Synthesizer:
    def __init__(self, config=None, lfs_atmo=None, lfs_nlte=None, dll=None):
        self.config, self.lfs_atmo, self.lfs_nlte = setup_lfs(
            config, lfs_atmo, lfs_nlte
        )
        # dict: internal storage of the adaptive wavelength grid
        self.wint = {}
        # dll: the smelib object used for the radiative transfer calculation
        self.dll = dll if dll is not None else SME_DLL()
        self.dll = self.get_dll_id()
        self.atmosphere_interpolator = None
        # This stores a reference to the currently used sme structure, so we only log it once
        self.known_sme = None
        logger.info("Don't forget to cite your sources. Use sme.citation()")

    def get_atmosphere(self, sme):
        """
        Return an atmosphere based on specification in an SME structure

        sme.atmo.method defines mode of action:
            "grid"
                interpolate on atmosphere grid
            "embedded"
                No change
            "routine"
                calls sme.atmo.source(sme, atmo)

        Parameters
        ---------
            sme : SME_Struct
                sme structure with sme.atmo = atmosphere specification

        Returns
        -------
        sme : SME_Struct
            sme structure with updated sme.atmo
        """

        # Handle atmosphere grid or user routine.
        atmo = sme.atmo

        if atmo.method == "grid":
            if self.atmosphere_interpolator is None:
                self.atmosphere_interpolator = AtmosphereInterpolator(
                    depth=atmo.depth,
                    interp=atmo.interp,
                    geom=atmo.geom,
                    lfs_atmo=self.lfs_atmo,
                )
            else:
                self.atmosphere_interpolator.depth = atmo.depth
                self.atmosphere_interpolator.interp = atmo.interp
                self.atmosphere_interpolator.geom = atmo.geom

            atmo = self.atmosphere_interpolator.interp_atmo_grid(
                atmo.source, sme.teff, sme.logg, sme.monh
            )
        elif atmo.method == "routine":
            atmo = atmo.source(sme, atmo)
        elif atmo.method == "embedded":
            # atmo structure already extracted in sme_main
            pass
        else:
            raise AttributeError("Source must be 'grid', 'routine', or 'embedded'")

        sme.atmo = atmo
        return sme

    @staticmethod
    def get_wavelengthrange(wran, vrad, vsini):
        """
        Determine wavelengthrange that needs to be calculated
        to include all lines within velocity shift vrad + vsini
        """
        # 30 km/s == maximum barycentric velocity
        vrad_pad = 30.0 + 0.5 * np.clip(vsini, 0, None)  # km/s
        vbeg = vrad_pad + np.clip(vrad, 0, None)  # km/s
        vend = vrad_pad - np.clip(vrad, None, 0)  # km/s

        wbeg = wran[0] * (1 - vbeg / clight)
        wend = wran[1] * (1 + vend / clight)
        return wbeg, wend

    @staticmethod
    def new_wavelength_grid(wint):
        """Generate new wavelength grid within bounds of wint"""
        # Determine step size for a new model wavelength scale, which must be uniform
        # to facilitate convolution with broadening kernels. The uniform step size
        # is the larger of:
        #
        # [1] smallest wavelength step in WINT_SEG, which has variable step size
        # [2] 10% the mean dispersion of WINT_SEG
        # [3] 0.05 km/s, which is 1% the width of solar line profiles

        wbeg, wend = wint[0], wint[-1]
        wmid = 0.5 * (wend + wbeg)  # midpoint of segment
        wspan = wend - wbeg  # width of segment
        diff = wint[1:] - wint[:-1]
        jmin = np.argmin(diff)
        vstep1 = diff[jmin] / wint[jmin] * clight  # smallest step
        vstep2 = 0.1 * wspan / (len(wint) - 1) / wmid * clight  # 10% mean dispersion
        vstep3 = 0.05  # 0.05 km/s step
        vstep = max(vstep1, vstep2, vstep3)  # select the largest

        # Generate model wavelength scale X, with uniform wavelength step.
        nx = int(
            np.abs(np.log10(wend / wbeg)) / np.log10(1 + vstep / clight) + 1
        )  # number of wavelengths
        if nx % 2 == 0:
            nx += 1  # force nx to be odd

        # Resolution
        # IDL way
        # resol_out = 1 / ((wend / wbeg) ** (1 / (nx - 1)) - 1)
        # vstep = clight / resol_out
        # x_seg = wbeg * (1 + 1 / resol_out) ** np.arange(nx)

        # Python way (not identical, as IDL endpoint != wend)
        # difference approx 1e-7
        x_seg = np.geomspace(wbeg, wend, num=nx)
        resol_out = 1 / np.diff(np.log(x_seg[:2]))[0]
        vstep = clight / resol_out
        return x_seg, vstep

    @staticmethod
    def check_segments(sme, segments):
        if isinstance(segments, str) and segments == "all":
            segments = range(sme.nseg)
        else:
            segments = np.atleast_1d(segments)
            if np.any(segments < 0) or np.any(segments >= sme.nseg):
                raise IndexError("Segment(s) out of range")
            segments = np.unique(segments)

        if sme.mask is not None:
            segments = [
                seg for seg in segments if not np.all(sme.mask[seg] == MASK_VALUES.BAD)
            ]
        return segments

    @staticmethod
    def apply_radial_velocity_and_continuum(
        wave, spec, wmod, smod, cmod, vrad, cscale, cscale_type, segments
    ):
        smod = apply_radial_velocity_and_continuum(
            wave, wmod, smod, vrad, cscale, cscale_type, segments
        )
        cmod = apply_radial_velocity_and_continuum(
            wave, wmod, cmod, vrad, None, None, segments
        )
        return smod, cmod

    @staticmethod
    def integrate_flux(mu, inten, deltav, vsini, vrt, osamp=None):
        """
        Produces a flux profile by integrating intensity profiles (sampled
        at various mu angles) over the visible stellar surface.

        Intensity profiles are weighted by the fraction of the projected
        stellar surface they represent, apportioning the area between
        adjacent MU points equally. Additional weights (such as those
        used in a Gauss-Legendre quadrature) can not meaningfully be
        used in this scheme.  About twice as many points are required
        with this scheme to achieve the precision of Gauss-Legendre
        quadrature.
        DELTAV, VSINI, and VRT must all be in the same units (e.g. km/s).
        If specified, OSAMP should be a positive integer.

        Parameters
        ----------
        mu : array(float) of size (nmu,)
            cosine of the angle between the outward normal and
            the line of sight for each intensity spectrum in INTEN.
        inten : array(float) of size(nmu, npts)
            intensity spectra at specified values of MU.
        deltav : float
            velocity spacing between adjacent spectrum points
            in INTEN (same units as VSINI and VRT).
        vsini : float
            maximum radial velocity, due to solid-body rotation.
        vrt : float
            radial-tangential macroturbulence parameter, i.e.
            np.sqrt(2) times the standard deviation of a Gaussian distribution
            of turbulent velocities. The same distribution function describes
            the radial motions of one component and the tangential motions of
            a second component. Each component covers half the stellar surface.
            See 'The Observation and Analysis of Stellar Photospheres', Gray.
        osamp : int, optional
            internal oversampling factor for convolutions.
            By default convolutions are done using the input points (OSAMP=1),
            but when OSAMP is set to higher integer values, the input spectra
            are first oversampled by cubic spline interpolation.

        Returns
        -------
        value : array(float) of size (npts,)
            Disk integrated flux profile.

        Note
        ------------
            If you use this algorithm in work that you publish, please cite
            Valenti & Anderson 1996, PASP, currently in preparation.
        """
        """
        History
        -----------
        Feb-88  GM
            Created ANA version.
        13-Oct-92 JAV
            Adapted from G. Marcy's ANA routi!= of the same name.
        03-Nov-93 JAV
            Switched to annular convolution technique.
        12-Nov-93 JAV
            Fixed bug. Intensity compo!=nts not added when vsini=0.
        14-Jun-94 JAV
            Reformatted for "public" release. Heavily commented.
            Pass deltav instead of 2.998d5/deltav. Added osamp
            keyword. Added rebinning logic at end of routine.
            Changed default osamp from 3 to 1.
        20-Feb-95 JAV
            Added mu as an argument to handle arbitrary mu sampling
            and remove ambiguity in intensity profile ordering.
            Interpret VTURB as np.sqrt(2)*sigma instead of just sigma.
            Replaced call_external with call to spl_{init|interp}.
        03-Apr-95 JAV
            Multiply flux by pi to give observed flux.
        24-Oct-95 JAV
            Force "nmk" padding to be at least 3 pixels.
        18-Dec-95 JAV
            Renamed from dskint() to rtint(). No longer make local
            copy of intensities. Use radial-tangential instead
            of isotropic Gaussian macroturbulence.
        26-Jan-99 JAV
            For NMU=1 and VSINI=0, assume resolved solar surface#
            apply R-T macro, but supress vsini broadening.
        01-Apr-99 GMH
            Use annuli weights, rather than assuming ==ual area.
        07-Mar-12 JAV
            Force vsini and vmac to be scalars.
        """

        # Make local copies of various input variables, which will be altered below.
        # Force vsini and especially vmac to be scalars. Otherwise mu dependence fails.

        if np.size(vsini) > 1:
            vsini = vsini[0]
        if np.size(vrt) > 1:
            vrt = vrt[0]
        nmu = np.size(mu)  # number of radii

        # Convert input MU to projected radii, R, of annuli for a star of unit radius
        #  (which is just sine, rather than cosine, of the angle between the outward
        #  normal and the line of sight).
        rmu = np.sqrt(1 - mu ** 2)  # use simple trig identity
        if nmu > 1:
            r = np.sqrt(
                0.5 * (rmu[:-1] ** 2 + rmu[1:] ** 2)
            )  # area midpoints between rmu
            r = np.concatenate(([0], r, [1]))
        else:
            r = np.array([0, 1])

        # Determine oversampling factor.
        if osamp is None:
            if vsini == 0:
                os = 2
            else:
                os = deltav / (vsini * r[r != 0])
                os = np.max(os[np.isfinite(os)])
                os = int(np.ceil(os)) + 1
        else:
            os = osamp
        # force integral value > 1
        os = round(np.clip(os, 2, 10))

        # Sort the projected radii and corresponding intensity spectra into ascending
        #  order (i.e. from disk center to the limb), which is equivalent to sorting
        #  MU in descending order.
        isort = np.argsort(rmu)
        rmu = rmu[isort]  # reorder projected radii
        if nmu == 1 and vsini != 0:
            logger.warning(
                "Vsini is non-zero, but only one projected radius (mu value) is set. No rotational broadening will be performed."
            )
            vsini = 0  # ignore vsini if only 1 mu

        # Calculate projected radii for boundaries of disk integration annuli.  The n+1
        # boundaries are selected such that r(i+1) exactly bisects the area between
        # rmu(i) and rmu(i+1). The in!=rmost boundary, r(0) is set to 0 (disk center)
        # and the outermost boundary, r(nmu) is set to 1 (limb).
        if nmu > 1:  # really want disk integration
            # Calculate integration weights for each disk integration annulus.  The weight
            # is just given by the relative area of each annulus, normalized such that
            # the sum of all weights is unity.  Weights for limb darkening are included
            # explicitly in the intensity profiles, so they aren't needed here.
            wt = r[1:] ** 2 - r[:-1] ** 2  # weights = relative areas
        else:
            wt = np.array([1.0])  # single mu value, full weight

        # Generate index vectors for input and oversampled points. Note that the
        # oversampled indicies are carefully chosen such that every "os" finely
        # sampled points fit exactly into one input bin. This makes it simple to
        # "integrate" the finely sampled points at the end of the routine.
        npts = inten.shape[1]  # number of points
        xpix = np.arange(npts, dtype=float)  # point indices
        nfine = os * npts  # number of oversampled points
        xfine = (0.5 / os) * (
            2 * np.arange(nfine, dtype=float) - os + 1
        )  # oversampled points indices

        # Loop through annuli, constructing and convolving with rotation kernels.

        yfine = np.empty(nfine)  # init oversampled intensities
        flux = np.zeros(nfine)  # init flux vector
        for imu in range(nmu):  # loop thru integration annuli

            #  Use external cubic spline routine (adapted from Numerical Recipes) to make
            #  an oversampled version of the intensity profile for the current annulus.
            ypix = inten[isort[imu]]  # extract intensity profile
            if os == 1:
                # just copy (use) original profile
                yfine = np.copy(ypix)
            else:
                # spline onto fine wavelength scale
                try:
                    yfine = interp1d(
                        xpix, ypix, kind="cubic", fill_value="extrapolate"
                    )(xfine)
                except ValueError:
                    yfine = interp1d(
                        xpix, ypix, kind="linear", fill_value="extrapolate"
                    )(xfine)

            # Construct the convolution kernel which describes the distribution of
            # rotational velocities present in the current annulus. The distribution has
            # been derived analytically for annuli of arbitrary thickness in a rigidly
            # rotating star. The kernel is constructed in two pieces: o!= piece for
            # radial velocities less than the maximum velocity along the inner edge of
            # the annulus, and one piece for velocities greater than this limit.
            if vsini > 0:
                # nontrivial case
                r1 = r[imu]  # inner edge of annulus
                r2 = r[imu + 1]  # outer edge of annulus
                dv = deltav / os  # oversampled velocity spacing
                maxv = vsini * r2  # maximum velocity in annulus
                nrk = 2 * int(maxv / dv) + 3  ## oversampled kernel point
                # velocity scale for kernel
                v = dv * (np.arange(nrk, dtype=float) - ((nrk - 1) / 2))
                rkern = np.zeros(nrk)  # init rotational kernel
                j1 = np.abs(v) < vsini * r1  # low velocity points
                rkern[j1] = np.sqrt((vsini * r2) ** 2 - v[j1] ** 2) - np.sqrt(
                    (vsini * r1) ** 2 - v[j1] ** 2
                )  # generate distribution

                j2 = (np.abs(v) >= vsini * r1) & (np.abs(v) <= vsini * r2)
                rkern[j2] = np.sqrt(
                    (vsini * r2) ** 2 - v[j2] ** 2
                )  # generate distribution

                rkern = rkern / np.sum(rkern)  # normalize kernel

                # Convolve the intensity profile with the rotational velocity kernel for this
                # annulus. Pad each end of the profile with as many points as are in the
                # convolution kernel. This reduces Fourier ringing.
                yfine = convolve(yfine, rkern, mode="nearest")

            # Calculate projected sigma for radial and tangential velocity distributions.
            muval = mu[isort[imu]]  # current value of mu
            sigma = os * vrt / np.sqrt(2) / deltav  # standard deviation in points
            sigr = sigma * muval  # reduce by current mu value
            sigt = sigma * np.sqrt(1.0 - muval ** 2)  # reduce by np.sqrt(1-mu**2)

            # Figure out how many points to use in macroturbulence kernel.
            nmk = int(10 * sigma)
            nmk = np.clip(nmk, 3, (nfine - 3) // 2)

            # Construct radial macroturbulence kernel with a sigma of mu*VRT/np.sqrt(2).
            if sigr > 0:
                xarg = np.linspace(-nmk, nmk, 2 * nmk + 1) / sigr
                xarg = np.clip(-0.5 * xarg ** 2, -20, None)
                mrkern = np.exp(xarg)  # compute the gaussian
                mrkern = mrkern / np.sum(mrkern)  # normalize the profile
            else:
                mrkern = np.zeros(2 * nmk + 1)  # init with 0d0
                mrkern[nmk] = 1.0  # delta function

            # Construct tangential kernel with a sigma of np.sqrt(1-mu**2)*VRT/np.sqrt(2).
            if sigt > 0:
                xarg = np.linspace(-nmk, nmk, 2 * nmk + 1) / sigt
                xarg = np.clip(-0.5 * xarg ** 2, -20, None)
                mtkern = np.exp(xarg)  # compute the gaussian
                mtkern = mtkern / np.sum(mtkern)  # normalize the profile
            else:
                mtkern = np.zeros(2 * nmk + 1)  # init with 0d0
                mtkern[nmk] = 1.0  # delta function

            # Sum the radial and tangential components, weighted by surface area.
            area_r = 0.5  # assume equal areas
            area_t = 0.5  # ar+at must equal 1
            mkern = area_r * mrkern + area_t * mtkern  # add both components

            # Convolve the total flux profiles, again padding the spectrum on both ends to
            # protect against Fourier ringing.
            yfine = convolve(yfine, mkern, mode="nearest")

            # Add contribution from current annulus to the running total.
            flux = flux + wt[imu] * yfine  # add profile to running total

        flux = np.reshape(flux, (npts, os))  # convert to an array
        flux = np.pi * np.sum(flux, axis=1) / os  # sum, normalize
        return flux

    def get_dll_id(self, dll=None):
        if dll is None:
            dll = self.dll
        if dll in __DLL_IDS__:
            dll_id = __DLL_IDS__[dll]
        elif dll in __DLL_DICT__:
            dll_id = dll
        else:
            dll_id = uuid.uuid4()
            __DLL_DICT__[dll_id] = dll
            __DLL_IDS__[dll] = dll_id
        return dll_id

    def get_dll(self, dll_id=None):
        if dll_id is None:
            dll_id = self.dll
        if dll_id in __DLL_DICT__:
            return __DLL_DICT__[dll_id]
        else:
            return dll_id

    def synthesize_spectrum(
        self,
        sme,
        segments="all",
        passLineList=True,
        passAtmosphere=True,
        passNLTE=True,
        updateStructure=True,
        updateLineList=False,
        reuse_wavelength_grid=False,
        radial_velocity_mode="robust",
        dll_id=None,
    ):
        """
        Calculate the synthetic spectrum based on the parameters passed in the SME structure
        The wavelength range of each segment is set in sme.wran
        The specific wavelength grid is given by sme.wave, or is generated on the fly if sme.wave is None

        Will try to fit radial velocity RV and continuum to observed spectrum, depending on vrad_flag and cscale_flag

        Other important fields:
        sme.iptype: instrument broadening type

        Parameters
        ----------
        sme : SME_Struct
            sme structure, with all necessary parameters for the calculation
        setLineList : bool, optional
            wether to pass the linelist to the c library (default: True)
        passAtmosphere : bool, optional
            wether to pass the atmosphere to the c library (default: True)
        passNLTE : bool, optional
            wether to pass NLTE departure coefficients to the c library (default: True)
        reuse_wavelength_grid : bool, optional
            wether to use sme.wint as the output grid of the function or create a new grid (default: False)

        Returns
        -------
        sme : SME_Struct
            same sme structure with synthetic spectrum in sme.smod
        """

        if sme is not self.known_sme:
            logger.debug("Synthesize spectrum")
            logger.debug("%s", sme)
            self.known_sme = sme

        # Define constants
        n_segments = sme.nseg
        cscale_degree = sme.cscale_degree

        # fix impossible input
        if "spec" not in sme or sme.spec is None:
            sme.vrad_flag = "none"
            sme.cscale_flag = "none"
        else:
            if "uncs" not in sme or sme.uncs is None:
                sme.uncs = np.ones(sme.spec.size)
            if "mask" not in sme or sme.mask is None:
                sme.mask = np.full(sme.spec.size, MASK_VALUES.LINE)
            for i in range(sme.nseg):
                mask = ~np.isfinite(sme.spec[i])
                mask |= sme.uncs[i] == 0
                sme.mask[i][mask] = MASK_VALUES.BAD

        if radial_velocity_mode != "robust" and (
            "cscale" not in sme or "vrad" not in sme
        ):
            radial_velocity_mode = "robust"

        segments = self.check_segments(sme, segments)

        # Prepare arrays
        vrad, _, cscale, _ = null_result(sme.nseg, sme.cscale_degree, sme.cscale_type)

        wave = [np.zeros(0) for _ in range(n_segments)]
        smod = [[] for _ in range(n_segments)]
        cmod = [[] for _ in range(n_segments)]
        wmod = [[] for _ in range(n_segments)]

        # If wavelengths are already defined use those as output
        if "wave" in sme:
            wave = [w for w in sme.wave]

        dll = self.get_dll(dll_id)

        # Input Model data to C library
        dll.SetLibraryPath()
        if passLineList:
            dll.InputLineList(sme.linelist)
        if hasattr(updateLineList, "__len__") and len(updateLineList) > 0:
            # TODO Currently Updates the whole linelist, could be improved to only change affected lines
            dll.UpdateLineList(sme.atomic, sme.species, updateLineList)
        if passAtmosphere:
            sme = self.get_atmosphere(sme)
            dll.InputModel(sme.teff, sme.logg, sme.vmic, sme.atmo)
            dll.InputAbund(sme.abund)
            dll.Ionization(0)
            dll.SetVWscale(sme.gam6)
            dll.SetH2broad(sme.h2broad)
        if passNLTE:
            sme.nlte.update_coefficients(sme, dll, self.lfs_nlte)

        # Loop over segments
        #   Input Wavelength range and Opacity
        #   Calculate spectral synthesis for each
        #   Interpolate onto geomspaced wavelength grid
        #   Apply instrumental and turbulence broadening

        for il in tqdm(
            segments, desc="Segments", leave=False, disable=~show_progress_bars
        ):
            wmod[il], smod[il], cmod[il] = self.synthesize_segment(
                sme,
                il,
                reuse_wavelength_grid,
                il != segments[0],
                dll_id=dll_id,
            )

        for il in segments:
            if "wave" not in sme or len(sme.wave[il]) == 0:
                # trim padding
                wbeg, wend = sme.wran[il]
                itrim = (wmod[il] > wbeg) & (wmod[il] < wend)
                # Force endpoints == wavelength range
                wave[il] = np.concatenate(([wbeg], wmod[il][itrim], [wend]))

        if sme.specific_intensities_only:
            return wmod, smod, cmod

        # Fit continuum and radial velocity
        # And interpolate the flux onto the wavelength grid
        if radial_velocity_mode == "robust":
            cscale, cscale_unc, vrad, vrad_unc = match_rv_continuum(
                sme, segments, wmod, smod
            )
            logger.debug("Radial velocity: %s", str(vrad))
            logger.debug("Continuum coefficients: %s", str(cscale))
        elif radial_velocity_mode == "fast":
            cscale, vrad = sme.cscale, sme.vrad
        else:
            raise ValueError("Radial Velocity mode not understood")

        smod, cmod = self.apply_radial_velocity_and_continuum(
            wave,
            sme.spec,
            wmod,
            smod,
            cmod,
            vrad,
            cscale,
            sme.cscale_type,
            segments,
        )

        # Merge all segments
        # if sme already has a wavelength this should be the same
        if updateStructure:
            if "wave" not in sme:
                # TODO: what if not all segments are there?
                sme.wave = wave
            if "synth" not in sme:
                sme.synth = smod
            if "cont" not in sme:
                sme.cont = cmod

            for s in segments:
                sme.wave[s] = wave[s]
                sme.synth[s] = smod[s]
                sme.cont[s] = cmod[s]

            if sme.cscale_type in ["spline", "spline+mask"]:
                sme.cscale = np.asarray(cscale)
                sme.cscale_unc = np.asarray(cscale_unc)
            elif sme.cscale_flag not in ["fix", "none"]:
                for s in np.arange(sme.nseg):
                    if s not in segments:
                        cscale[s] = sme.cscale[s]
                sme.cscale = np.asarray(cscale)
                sme.cscale_unc = np.asarray(cscale_unc)

            sme.vrad = np.asarray(vrad)
            sme.vrad_unc = np.asarray(vrad_unc)
            sme.nlte.flags = dll.GetNLTEflags()
            result = sme
        else:
            wave = Iliffe_vector(values=wave)
            smod = Iliffe_vector(values=smod)
            cmod = Iliffe_vector(values=cmod)
            result = wave, smod, cmod

        # Cleanup
        return result

    def synthesize_segment(
        self,
        sme,
        segment,
        reuse_wavelength_grid=False,
        keep_line_opacity=False,
        dll_id=None,
    ):
        """Create the synthetic spectrum of a single segment

        Parameters
        ----------
        sme : SME_Struct
            The SME strcuture containing all relevant parameters
        segment : int
            the segment to synthesize
        reuse_wavelength_grid : bool
            Whether to keep the current wavelength grid for the synthesis
            or create a new one, depending on the linelist. Default: False
        keep_line_opacity : bool
            Whether to reuse existing line opacities or not. This should be
            True iff the opacities have been calculated in another segment.

        Returns
        -------
        wgrid : array of shape (npoints,)
            Wavelength grid of the synthesized spectrum
        flux : array of shape (npoints,)
            The Flux of the synthesized spectrum
        cont_flux : array of shape (npoints,)
            The continuum Flux of the synthesized spectrum
        """
        logger.debug("Segment %i out of %i", segment, sme.nseg)
        dll = self.get_dll(dll_id)

        # Input Wavelength range and Opacity
        vrad_seg = sme.vrad[segment] if sme.vrad[segment] is not None else 0
        wbeg, wend = self.get_wavelengthrange(sme.wran[segment], vrad_seg, sme.vsini)

        dll.InputWaveRange(wbeg, wend)
        dll.Opacity()

        # Reuse adaptive wavelength grid in the jacobians
        if reuse_wavelength_grid and segment in self.wint.keys():
            wint_seg = self.wint[segment]
        else:
            wint_seg = None

        # Only calculate line opacities in the first segment
        #   Calculate spectral synthesis for each
        _, wint, sint, cint = dll.Transf(
            sme.mu,
            accrt=sme.accrt,  # threshold line opacity / cont opacity
            accwi=sme.accwi,
            keep_lineop=keep_line_opacity,
            wave=wint_seg,
        )

        # Store the adaptive wavelength grid for the future
        # if it was newly created
        if wint_seg is None:
            self.wint[segment] = wint

        if not sme.specific_intensities_only:
            # Create new geomspaced wavelength grid, to be used for intermediary steps
            wgrid, vstep = self.new_wavelength_grid(wint)

            logger.debug("Integrate specific intensities")
            # Radiative Transfer Integration
            # Continuum
            cint = self.integrate_flux(sme.mu, cint, 1, 0, 0)
            cint = np.interp(wgrid, wint, cint)

            # Broaden Spectrum
            y_integrated = np.empty((sme.nmu, len(wgrid)))
            for imu in range(sme.nmu):
                y_integrated[imu] = np.interp(wgrid, wint, sint[imu])

            # Turbulence broadening
            # Apply macroturbulent and rotational broadening while integrating intensities
            # over the stellar disk to produce flux spectrum Y.
            sint = self.integrate_flux(sme.mu, y_integrated, vstep, sme.vsini, sme.vmac)
            wint = wgrid

            # instrument broadening
            if "iptype" in sme:
                logger.debug("Apply detector broadening")
                ipres = sme.ipres if np.size(sme.ipres) == 1 else sme.ipres[segment]
                sint = broadening.apply_broadening(
                    ipres, wint, sint, type=sme.iptype, sme=sme
                )

        # Divide calculated spectrum by continuum
        if sme.normalize_by_continuum:
            sint /= cint

        return wint, sint, cint


def synthesize_spectrum(sme, segments="all"):
    synthesizer = Synthesizer()
    return synthesizer.synthesize_spectrum(sme, segments)



# -*- coding: utf-8 -*-
import logging

import numpy as np
from scipy.optimize import curve_fit

logger = logging.getLogger(__name__)


def uncertainties(pder, resid, unc, freep_name, plot=False):
    binsize = np.median(np.abs(resid))
    nbins = (np.max(resid) - np.min(resid)) / binsize + 1
    nbins = int(nbins)
    h, x = np.histogram(resid, bins=nbins)
    x = x[:-1] + np.diff(x) / 2

    _, a = gaussfit(x, h, nterms="constant")
    sigma = a[2]

    # 6 sigma threshold
    resid_threshold = 6 * sigma

    nfree = len(freep_name)
    freep_unc = np.zeros(nfree)
    freep_med = np.zeros(nfree)
    freep_msig = np.zeros(nfree)
    freep_psig = np.zeros(nfree)

    for ifree in range(nfree):
        gradlim = np.median(np.abs(pder[:, ifree]))
        i = np.where(
            (np.abs(pder[:, ifree]) > gradlim) & (np.abs(resid) < resid_threshold)
        )[0]
        ni = len(i)
        if ni > 21:
            ii = np.argsort(resid[i] / pder[i, ifree])
            # Sort pixels according to the change of the ifree
            # parameter needed to match the observations
            ch_x = resid[i[ii]] / pder[i[ii], ifree]
            # Weights of the individual pixels also sorted
            ch_y = np.abs(pder[i[ii], ifree]) / unc[i[ii]]
            # Cumulative weights
            ch_y = np.cumsum(ch_y)
            # Normalized cumulative weights
            ch_y = ch_y / ch_y[-1]

            # value of distribution at -sigma
            freep_msig[ifree] = np.interp(0.5 - 0.6827 / 2, ch_x, ch_y)
            # value of distribution at +sigma
            freep_psig[ifree] = np.interp(0.5 + 0.6827 / 2, ch_x, ch_y)
            # Mean sigma
            freep_unc[ifree] = (freep_psig[ifree] - freep_msig[ifree]) * 0.5
            # Median change
            freep_med[ifree] = np.median(ch_y)

        print(
            "%s : %.3f %.3f %.3f %.3f"
            % (
                freep_name[ifree],
                freep_med[ifree],
                freep_msig[ifree],
                freep_psig[ifree],
                freep_unc[ifree],
            )
        )

        if plot:
            # TODO plot
            pass

    return freep_med, freep_msig, freep_psig, freep_unc


def gaussfit(x, y, nterms="none"):
    """
    Fit a simple gaussian to data

    gauss(x, a, mu, sigma) = a * exp(-z**2/2)
    with z = (x - mu) / sigma

    Parameters
    ----------
    x : array(float)
        x values
    y : array(float)
        y values
    Returns
    -------
    gauss(x), parameters
        fitted values for x, fit paramters (a, mu, sigma)
    """

    z = lambda x, A1, A2: (x - A1) / A2

    if nterms in ["none", 3]:
        gauss = lambda x, A0, A1, A2: A0 * np.exp(-z(x, A1, A2) ** 2 / 2)
        p0 = [max(y), 1, 1]
    elif nterms in ["constant", 4]:
        gauss = lambda x, A0, A1, A2, A3: A0 * np.exp(-z(x, A1, A2) ** 2 / 2) + A3
        p0 = [max(y), 1, 1, 0]
    elif nterms in ["linear", 5]:
        gauss = (
            lambda x, A0, A1, A2, A3, A4: A0 * np.exp(-z(x, A1, A2) ** 2 / 2)
            + A3
            + A4 * x
        )
        p0 = [max(y), 1, 1, 0, 0]

    elif nterms in ["quadratic", 6]:
        gauss = (
            lambda x, A0, A1, A2, A3, A4, A5: A0 * np.exp(-z(x, A1, A2) ** 2 / 2)
            + A3
            + A4 * x
            + A5 * x ** 2
        )
        p0 = [max(y), 1, 1, 0, 0, 0]

    popt, _ = curve_fit(gauss, x, y, p0=p0)
    return gauss(x, *popt), popt



# -*- coding: utf-8 -*-
import argparse

from .gui import plot_plotly
from .sme import SME_Structure
from .solve import SME_Solver
from .synthesize import Synthesizer

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "PySME",
        description=(
            "Synthesizes stellar spectra and determines best fit parameters "
            "to observations. The results are stored in the same file as the input."
        ),
    )
    parser.add_argument("file", help=".sme input file")
    parser.add_argument(
        "-s",
        "--synthesize",
        action="store_true",
        help="Only synthesize the spectrum, no fitting",
    )
    parser.add_argument("-o", "--output", help="Store the output to this file instead")
    parser.add_argument(
        "-p",
        "--plot",
        nargs="?",
        const=True,
        default=False,
        help="Create a plot with the results",
    )
    args = parser.parse_args()
    print(args)

    synthesize_only = args.synthesize
    filename = args.file
    output = args.output
    plot = args.plot
    if output is None:
        output = filename

    sme = SME_Structure.load(filename)
    if synthesize_only:
        syn = Synthesizer()
        sme = syn.synthesize_spectrum(sme)
    else:
        solver = SME_Solver()
        sme = solver.solve(sme, sme.fitparameters)
    sme.save(output)

    if plot:
        fig = plot_plotly.FinalPlot(sme)
        if plot is True:
            fig.show()
        else:
            fig.save(filename=plot, auto_open=False)



# -*- coding: utf-8 -*-
import pytest
import requests

from pysme.config import Config
from pysme.large_file_storage import setup_atmo, setup_nlte


def lfs_available():
    config = Config()
    r = requests.head(config["data.file_server"])
    return r.status_code == 200


skipif_lfs = pytest.mark.skipif(lfs_available(), reason="LFS not available")


@pytest.fixture
def lfs_nlte():
    lfs_nlte = setup_nlte()
    yield lfs_nlte


@pytest.fixture
def lfs_atmo():
    lfs_atmo = setup_atmo()
    yield lfs_atmo



# -*- coding: utf-8 -*-
# Get and read the NSO spectrum
import numpy as np

from . import large_file_storage as lfs


def load_solar_spectrum(ravel=True):
    """
    Load the solar spectrum

    The data is from the National Solar Observatory Atlas #1, which is an
    optical, full-disk, FTS spectrum of a relatively inactive Sun.
    Automatically retrieves the data from the SME Large File Server
    and stores it for local use.

    Returns
    -------
    wave : array
        Wavelength in Angström
    flux : array
        Normalized spectrum
    """
    server = lfs.setup_atmo()
    fname = server.get("nso.bin")

    # number of wavelength segments
    nseg = 1073
    # spectrum points per segment
    pts = 1024
    # the data is split into wavelength and flux
    # as defined here
    dtype = [("wave", f"({pts},)>f4"), ("flux", f"({pts},)>i2")]

    with open(fname, "rb") as f:
        # Read the wavelength of each segment from the header
        wseg = np.fromfile(f, dtype=">i2", count=nseg)
        data = np.fromfile(f, dtype=dtype)

    # Split the data
    wave = data["wave"]
    flux = data["flux"]
    # Add the base wavelength to each segment
    wave = wave + wseg[:, None]
    # Normalize spectrum (and change to double data type implicitly)
    flux = flux / 30_000

    # Unravel the spectra
    if ravel:
        wave = wave.ravel()
        flux = flux.ravel()

    return wave, flux



# -*- coding: utf-8 -*-
"""
These libtools are used to locate, download, and install the SME C/Fortran
library.
"""

import ctypes as ct
import logging
import os
import platform
import subprocess
import sys
import zipfile
from os.path import basename, dirname, exists, join, realpath

import wget

logger = logging.getLogger(__name__)


def download_libsme(loc=None):
    """
    Download the SME library and the necessary datafiles

    Parameters
    ----------
    loc : str, optional
        the path to the location the files should be placed,
        by default they are placed so that PySME can find and use them

    Raises
    ------
    KeyError
        If no existing library is found for this system
    """
    if loc is None:
        loc = dirname(dirname(get_full_libfile()))
    # Download compiled library from github releases
    print("Downloading and installing the latest libsme version for this system")
    aliases = {
        "Linux": "manylinux2014_x86_64",
        "Windows": "windows",
        "Darwin": "macos",
    }
    system = platform.system()

    try:
        system = aliases[system]
        print("Identified OS: %s" % system)
    except KeyError:
        raise KeyError(
            f"Could not find the associated compiled library for this system {system}."
            " Either compile it yourself and place it in src/pysme/ or open an"
            " issue on Github. Supported systems are: Linux, MacOS, Windows."
        )

    github_releases_url = "https://github.com/AWehrhahn/SMElib/releases/latest/download"
    github_releases_fname = "{system}-gfortran.zip".format(system=system)
    url = github_releases_url + "/" + github_releases_fname
    fname = join(loc, github_releases_fname)

    try:
        os.remove(fname)
    except FileNotFoundError:
        pass

    print("Downloading file %s" % url)
    os.makedirs(loc, exist_ok=True)
    wget.download(url, out=loc)
    # the wget progress bar, does not include a new line
    print("")

    print("Extracting file")
    zipfile.ZipFile(fname).extractall(loc)

    try:
        os.remove(fname)
    except FileNotFoundError:
        pass

    print("done")

    if system in ["macos"]:
        # Need to adjust the install_names in the dylib
        print("Fixing the file paths in the .dylib file")
        fname = realpath(get_full_libfile())
        subprocess.run(
            ["install_name_tool", "-id", fname, fname], capture_output=True, check=True
        )


def compile_interface():
    """
    Compiles the Python Module Interface to the SME library. This needs to be
    called once, before trying to import _smelib.

    Since the module uses the setup.py method to be compiled,
    it is somewhat hacked together to make it work.
    """
    libdir = join(dirname(__file__))
    executable = sys.executable
    if executable is None:
        # If python is unable to identify the path to its own executable use python3
        # This is unlikely to happen for us though
        executable = "python3"
    cwd = os.getcwd()
    # We need to swith to the correct directory and back, for setup.py to work
    os.chdir(libdir)
    subprocess.run([executable, "setup.py", "build_ext", "--inplace"])
    os.chdir(cwd)


def get_lib_name():
    """Get the name of the SME C library"""
    system = platform.system().lower()

    if system == "windows":
        return "libsme-5.dll"
    elif system == "darwin":
        return "libsme.dylib"

    arch = platform.machine()
    bits = 64  # platform.architecture()[0][:-3]

    return "sme_synth.so.{system}.{arch}.{bits}".format(
        system=system, arch=arch, bits=bits
    )


def get_lib_directory():
    """
    Get the directory name of the library. I.e. 'lib' on all systems
    execpt windows, and 'bin' on windows
    """
    if platform.system() in ["Windows"]:
        dirpath = "bin"
    else:
        # For Linux/MacOS
        dirpath = "lib"
    return dirpath


def get_full_libfile():
    """Get the full path to the sme C library"""
    localdir = dirname(dirname(__file__))
    libfile = get_lib_name()
    dirpath = get_lib_directory()
    libfile = join(localdir, dirpath, libfile)
    return libfile


def load_library(libfile=None):
    """
    Load the SME library using cytpes.CDLL

    This is useful and necessary for the pymodule interface to find the
    library.

    Parameters
    ----------
    libfile : str, optional
        filename of the library to load, by default use the SME library in
        this package

    Returns
    -------
    lib : CDLL
        library object of the SME library
    """
    if libfile is None:
        libfile = get_full_libfile()
    try:
        os.add_dll_directory(dirname(libfile))
    except AttributeError:
        newpath = dirname(libfile)
        if "PATH" in os.environ:
            newpath += os.pathsep + os.environ["PATH"]
        os.environ["PATH"] = newpath
    return ct.CDLL(str(libfile))


def get_full_datadir():
    """
    Get the filepath to the datafiles of the SME library
    """
    localdir = dirname(dirname(__file__))
    datadir = join(localdir, "share/libsme/")
    return datadir



# -*- coding: utf-8 -*-
import io
import json
import logging
import os
import subprocess
import sys
import tempfile
from zipfile import ZIP_LZMA, ZIP_STORED, ZipFile

import numpy as np
from flex.flex import FlexExtension, FlexFile

from . import __version__

logger = logging.getLogger(__name__)


def to_flex(sme):
    header = {}
    extensions = {}

    for name in sme._names:
        value = sme[name]
        if isinstance(value, IPersist):
            extensions[name] = value._save()
        elif isinstance(value, FlexExtension):
            extensions[name] = value
        elif value is not None:
            header[name] = value

    ff = FlexFile(header, extensions)
    return ff


def from_flex(ff, sme):
    header = ff.header
    extensions = ff.extensions
    for name in sme._names:
        if name in updates.keys():
            name = updates[name]
        if name in header.keys():
            sme[name] = header[name]
        elif name in extensions.keys():
            if sme[name] is not None and isinstance(sme[name], IPersist):
                sme[name] = sme[name]._load(extensions[name])
            else:
                sme[name] = extensions[name]
    return sme


def save(filename, sme, format="flex", _async=False):
    """
    Create a folder structure inside a tarfile
    See flex-format for details

    Parameters
    ----------
    filename : str
        Filename of the final file
    sme : SME_Structure
        sme structure to save
    compressed : bool, optional
        whether to compress the output
    """
    ff = to_flex(sme)

    if format == "flex":
        file_ending = ".sme"
    else:
        file_ending = "." + format
    if not filename.endswith(file_ending):
        filename = filename + file_ending

    if format == "flex":
        if _async:
            ff.write_async(filename)
        else:
            ff.write(filename)
    elif format == "fits":
        ff.to_fits(filename, overwrite=True)
    elif format == "json":
        ff.to_json(filename)
    else:
        raise ValueError(
            "Format {!r} not understood, expected one of ['flex', 'fits', 'json'].".format(
                format
            )
        )


def load(fname, sme):
    """
    Load the SME Structure from disk

    Parameters
    ----------
    fname : str
        file to load
    sme : SME_Structure
        empty sme structure with default values set

    Returns
    -------
    sme :  SME_Structure
        loaded sme structure
    """
    try:
        ff = FlexFile.read(fname)
        sme = from_flex(ff, sme)
        ff.close()
        return sme
    except Exception as ex:
        logger.error(ex)
        try:
            sme = load_v1(fname, sme)
        except:
            raise ex
        return sme


# Update this if the names in sme change
updates = {"idlver": "system_info"}


class IPersist:
    def _save(self):
        raise NotImplementedError

    @classmethod
    def _load(cls, ext):
        raise NotImplementedError

    def _save_v1(self, file, folder=""):
        saves_v1(file, self, folder)

    @classmethod
    def _load_v1(cls, file, names, folder=""):
        logger.setLevel(logging.INFO)
        data = cls()  # TODO Suppress warnings
        data = loads_v1(file, data, names, folder)
        logger.setLevel(logging.NOTSET)
        return data


# Version 1 IO (Deprecated)


def toBaseType(value):
    if value is None:
        return value
    if isinstance(value, np.ndarray):
        return value.tolist()
    if isinstance(value, np.integer):
        return int(value)
    if isinstance(value, np.floating):
        return float(value)
    if isinstance(value, np.bool_):
        return bool(value)
    if isinstance(value, np.str):
        return str(value)

    return value


def save_v1(filename, data, folder="", compressed=False):
    """
    Create a folder structure inside a zipfile
    Add .json and .npy and .npz files with the correct names
    And subfolders for more complicated objects
    with the same layout
    Each class should have a save and a load method
    which can be used for this purpose

    Parameters
    ----------
    filename : str
        Filename of the final zipfile
    data : SME_struct
        data to save
    folder : str, optional
        subfolder to save data to
    compressed : bool, optional
        whether to compress the output
    """
    # We use LZMA for compression, since that yields the
    # smallest filesize of the existing compression algorithms
    if not compressed:
        compression = ZIP_STORED
    else:
        compression = ZIP_LZMA

    with ZipFile(filename, "w", compression) as file:
        saves_v1(file, data, folder=folder)


# TODO: this is specific for Collection type objects
# Move this to Collection, and not here
def saves_v1(file, data, folder=""):
    if folder != "" and folder[-1] != "/":
        folder = folder + "/"

    parameters = {}
    arrays = {}
    others = {}
    for key in data._names:
        value = getattr(data, key)
        if np.isscalar(value) or isinstance(value, dict):
            parameters[key] = value
        elif isinstance(value, (list, np.ndarray)):
            if np.size(value) > 20:
                arrays[key] = value
            else:
                parameters[key] = value
        else:
            others[key] = value

    info = json.dumps(parameters, default=toBaseType)
    file.writestr(f"{folder}info.json", info)

    for key, value in arrays.items():
        b = io.BytesIO()
        np.save(b, value)
        file.writestr(f"{folder}{key}.npy", b.getvalue())

    for key, value in others.items():
        if value is not None:
            value._save_v1(file, f"{folder}{key}")


def load_v1(filename, data):
    with ZipFile(filename, "r") as file:
        names = file.namelist()
        return loads_v1(file, data, names)


def loads_v1(file, data, names=None, folder=""):
    if folder != "" and folder[-1] != "/":
        folder = folder + "/"
    if names is None:
        names = file.namelist()

    subdirs = {}
    local = []
    for name in names:
        name_within = name[len(folder) :]
        if "/" not in name_within:
            local.append(name)
        else:
            direc, _ = name_within.split("/", 1)
            if direc not in subdirs.keys():
                subdirs[direc] = []
            subdirs[direc].append(name)

    for name in local:
        if name.endswith(".json"):
            info = file.read(name)
            info = json.loads(info)
            for key, value in info.items():
                key = updates.get(key, key)
                data[key] = value
        elif name.endswith(".npy"):
            b = io.BytesIO(file.read(name))
            key = name[len(folder) : -4]
            key = updates.get(key, key)
            data[key] = np.load(b)
        elif name.endswith(".npz"):
            b = io.BytesIO(file.read(name))
            key = name[len(folder) : -4]
            key = updates.get(key, key)
            value = np.load(b)
            data[key] = [value[f"arr_{i}"] for i in range(len(value))]

    for key, value in subdirs.items():
        data_key = updates.get(key, key)
        data[data_key] = data[data_key]._load_v1(file, value, folder=folder + key)

    return data


# IDL IO


def get_typecode(dtype):
    """Get the IDL typecode for a given dtype"""
    if dtype.name[:5] == "bytes":
        return "1"
    if dtype.name == "int16":
        return "2"
    if dtype.name == "int32":
        return "3"
    if dtype.name == "float32":
        return "4"
    if dtype.name == "float64":
        return "5"
    if dtype.name[:3] == "str":
        return dtype.name[3:]
    raise ValueError("Don't recognise the datatype")


temps_to_clean = []


def save_as_binary(arr):
    global temps_to_clean

    with tempfile.NamedTemporaryFile("w+", suffix=".dat", delete=False) as temp:
        if arr.dtype.name[:3] == "str" or arr.dtype.name == "object":
            arr = arr.astype(bytes)
            shape = (arr.dtype.itemsize, len(arr))
        elif np.issubdtype(arr.dtype, np.floating):
            # SME expects double precision, so we assure that here
            arr = arr.astype("float64")
            shape = arr.shape[::-1]
        else:
            shape = arr.shape[::-1]

        # Most arrays should be in the native endianness anyway
        # But if not we swap it to the native representation
        endian = arr.dtype.str[0]
        if endian == "<":
            endian = "little"
        elif endian == ">":
            endian = "big"
        elif endian == "|":
            endian = sys.byteorder

        if endian != sys.byteorder:
            arr = arr.newbyteorder().byteswap()
            endian = "native"

        arr.tofile(temp)
        value = [temp.name, str(list(shape)), get_typecode(arr.dtype), endian]
    temps_to_clean += [temp]
    return value


def clean_temps():
    global temps_to_clean
    for temp in temps_to_clean:
        try:
            os.remove(temp)
        except:
            pass

    temps_to_clean = []


def write_as_idl(sme):
    """
    Write SME structure into and idl format
    data arrays are stored in seperate temp files, and only the filename is passed to idl
    """

    vrad_flag = {"none": -2, "whole": -1, "each": 0, "fix": -2}[sme.vrad_flag]
    # cscale_flag = {"none": -3, "fix": -3, "constant": 0, "linear": 1, "quadratic": 1, }[
    #     sme.cscale_flag
    # ]
    # if not sme.normalize_by_continuum:
    #     cscale_flag = -2

    abund = sme.abund.get_pattern(type="sme", raw=True)
    abund[np.isnan(abund)] = -99

    fitvars = ["TEFF", "GRAV", "FEH", "VMIC", "VMAC", "VSINI", "GAM6", "VRAD"]
    fitvars = [s.upper() for s in sme.fitparameters if s.upper() in fitvars]
    if "logg" in sme.fitparameters:
        fitvars += ["GRAV"]
    if "monh" in sme.fitparameters:
        fitvars += ["FEH"]

    if sme.mask is None and sme.wave is not None:
        sme.mask = 1

    idl_fields = {
        "version": 5.1,
        "id": sme.id,
        "teff": sme.teff,
        "grav": sme.logg,
        "feh": sme.monh,
        "vmic": float(sme.vmic),
        "vmac": float(sme.vmac),
        "vsini": float(sme.vsini),
        "vrad": sme.vrad.tolist() if vrad_flag == 0 else sme.vrad[0],
        "vrad_flag": vrad_flag,
        "cscale": 1.0,
        "cscale_flag": 0,
        "gam6": sme.gam6,
        "h2broad": int(sme.h2broad),
        "accwi": sme.accwi,
        "accrt": sme.accrt,
        "clim": 0.01,
        "maxiter": 100,
        "chirat": 0.002,
        "nmu": sme.nmu,
        "nseg": sme.nseg,
        "abund": save_as_binary(abund),
        "species": save_as_binary(sme.species),
        "atomic": save_as_binary(sme.atomic),
        "lande": save_as_binary(sme.linelist.lande),
        "lineref": save_as_binary(sme.linelist.reference),
        "short_line_format": {"short": 1, "long": 2}[sme.linelist.lineformat],
        "wran": sme.wran.tolist(),
        "mu": sme.mu.tolist() if sme.nmu > 1 else sme.mu[0],
        "obs_name": "",
        "obs_type": 0,
        "glob_free": fitvars if len(fitvars) != 0 else "",
        "atmo": {
            "method": str(sme.atmo.method),
            "source": str(sme.atmo.source),
            "depth": str(sme.atmo.depth),
            "interp": str(sme.atmo.interp),
            "geom": str(sme.atmo.geom),
        },
    }

    if len(sme.nlte.elements) != 0:
        idl_fields["nlte"] = {}

        flags = np.zeros(99, dtype="int16")
        grids = ["" for _ in range(99)]
        for elem in sme.nlte.elements:
            flags[sme.abund.elem_dict[elem]] = 1
            grids[sme.abund.elem_dict[elem]] = sme.nlte.grids[elem]

        idl_fields["nlte"]["nlte_elem_flags"] = save_as_binary(flags)
        idl_fields["nlte"]["nlte_subgrid_size"] = save_as_binary(
            sme.nlte.subgrid_size.astype("int16")
        )
        idl_fields["nlte"]["nlte_grids"] = grids
        idl_fields["nlte"]["nlte_pro"] = "sme_nlte"

    if sme.iptype is not None:
        idl_fields["iptype"] = sme.iptype
        idl_fields["ipres"] = sme.ipres[0]
        # "ip_x": sme.ip_x,
        # "ip_y": sme.ip_y,
    else:
        idl_fields["iptype"] = "gauss"
        idl_fields["ipres"] = 0

    if sme.wave is not None:
        wind = np.cumsum(sme.wave.shape[1]) - 1
        idl_fields["wave"] = save_as_binary(sme.wave.ravel())
        idl_fields["wind"] = wind.tolist()
    if sme.spec is not None:
        idl_fields["sob"] = save_as_binary(sme.spec.ravel())
    if sme.uncs is not None:
        idl_fields["uob"] = save_as_binary(sme.uncs.ravel())
    if sme.mask is not None:
        idl_fields["mob"] = save_as_binary(sme.mask.ravel().astype("int16"))
    if sme.synth is not None:
        idl_fields["smod"] = save_as_binary(sme.synth.ravel())

    if "depth" in sme.linelist.columns:
        idl_fields["depth"] = save_as_binary(sme.linelist.depth)
    else:
        idl_fields["depth"] = save_as_binary(np.ones(len(sme.linelist)))

    if sme.linelist.lineformat == "long":
        idl_fields.update(
            {
                "line_extra": save_as_binary(sme.linelist.extra),
                "line_lulande": save_as_binary(sme.linelist.lulande),
                "line_term_low": save_as_binary(sme.linelist.term_lower),
                "line_term_upp": save_as_binary(sme.linelist.term_upper),
            }
        )

    sep = ""
    text = ""

    for key, value in idl_fields.items():
        if isinstance(value, dict):
            text += f"{sep}{key!s}:{{{key!s},$\n"
            sep = ""
            for key2, value2 in value.items():
                text += f"{sep}{key2!s}:{value2!r}$\n"
                sep = ","
            sep = ","
            text += "}$\n"
        else:
            text += f"{sep}{key!s}:{value!r}$\n"
            sep = ","
    return text


def save_as_idl(sme, fname):
    """
    Save the SME structure to disk as an idl save file

    This writes a IDL script to a temporary file, which is then run
    with idl as a seperate process. Therefore this reqires a working
    idl installation.

    There are two steps to this. First all the fields from the sme,
    structure need to be transformed into simple idl readable structures.
    All large arrays are stored in seperate binary files, for performance.
    The script then reads those files back into idl.
    """
    with tempfile.NamedTemporaryFile("w+", suffix=".pro") as temp:
        tempname = temp.name
        temp.write("print, 'Hello'\n")
        temp.write("sme = {sme,")
        # TODO: Save data as idl compatible data
        temp.write(write_as_idl(sme))
        temp.write("} \n")
        # This is the code that will be run in idl
        temp.write("print, 'there'\n")
        temp.write(
            """tags = tag_names(sme)
print, tags
new_sme = {}

for i = 0, n_elements(tags)-1 do begin
    arr = sme.(i)
    s = size(arr)
    if (s[0] eq 1) and (s[1] eq 4) then begin
        void = execute('shape = ' + arr[1])
        type = fix(arr[2])
        endian = string(arr[3])
        arr = read_binary(arr[0], data_dims=shape, data_type=type, endian=endian)
        if type eq 1 then begin
            ;string
            arr = string(arr)
        endif
    endif
    if (s[s[0]+1] eq 8) then begin
        ;struct
        tags2 = tag_names(sme.(i))
        new2 = {}
        tmp = sme.(i)

        for j = 0, n_elements(tags2)-1 do begin
            arr2 = tmp.(j)
            s = size(arr2)
            if (s[0] eq 1) and (s[1] eq 4) then begin
                void = execute('shape = ' + arr2[1])
                type = fix(arr2[2])
                endian = string(arr2[3])
                arr2 = read_binary(arr2[0], data_dims=shape, data_type=type, endian=endian)
                if type eq 1 then begin
                    ;string
                    arr2 = string(arr2)
                endif
            endif
            new2 = create_struct(temporary(new2), tags2[j], arr2)
        endfor
        arr = new2
    endif
    new_sme = create_struct(temporary(new_sme), tags[i], arr)
endfor

sme = new_sme\n"""
        )
        temp.write(f'save, sme, filename="{fname}"\n')
        temp.write("end\n")
        temp.flush()

        # with open(os.devnull, 'w') as devnull:
        print("IDL Script: ", tempname)
        subprocess.run(["idl", "-e", ".r %s" % tempname])
        # input("Wait for me...")
        clean_temps()



# -*- coding: utf-8 -*-
import os
from os.path import basename
from tempfile import NamedTemporaryFile

import numpy as np
from scipy.io import readsav

from .atmosphere import AtmosphereGrid


class SavFile(AtmosphereGrid):
    """IDL savefile atmosphere grid"""

    _cache = {}

    def __new__(cls, filename, source=None, lfs=None):
        # Check if the file is in the cache
        try:
            return cls._cache[filename]
        except:
            pass
        # Try loading the datafile using Numpy which is faster
        # and was generated in a previous iteration of PySME
        try:
            self = cls.load(filename)
            cls._cache = self
            return self
        except:
            pass

        # Otherwise we parse the sav file
        data = readsav(filename)

        npoints = data["atmo_grid_maxdep"]
        ngrids = data["atmo_grid_natmo"]
        self = super(SavFile, cls).__new__(cls, ngrids, npoints)
        if source is None:
            source = basename(filename)
        self.source = source

        try:
            self.info = b" ".join(data["atmo_grid_intro"]).decode()
        except (KeyError, TypeError):
            self.info = ""

        # TODO cover all cases
        if "marcs" in self.source:
            self.citation_info = r"""
                @ARTICLE{2008A&A...486..951G,
                    author = {{Gustafsson}, B. and {Edvardsson}, B. and {Eriksson}, K. and
                    {J{\o}rgensen}, U.~G. and {Nordlund}, {\r{A}}. and {Plez}, B.},
                    title = "{A grid of MARCS model atmospheres for late-type stars. I. Methods and general properties}",
                    journal = {Astronomy and Astrophysics},
                    keywords = {stars: atmospheres, Sun: abundances, stars: fundamental parameters, stars: general, stars: late-type, stars: supergiants, Astrophysics},
                    year = "2008",
                    month = "Aug",
                    volume = {486},
                    number = {3},
                    pages = {951-970},
                    doi = {10.1051/0004-6361:200809724},
                    archivePrefix = {arXiv},
                    eprint = {0805.0554},
                    primaryClass = {astro-ph},
                    adsurl = {https://ui.adsabs.harvard.edu/abs/2008A&A...486..951G},
                    adsnote = {Provided by the SAO/NASA Astrophysics Data System}
                }
            """
        elif "atlas" in self.source:
            self.citation_info = r"""
                @MISC{2017ascl.soft10017K,
                    author = {{Kurucz}, Robert L.},
                    title = "{ATLAS9: Model atmosphere program with opacity distribution functions}",
                    keywords = {Software},
                    year = "2017",
                    month = "Oct",
                    eid = {ascl:1710.017},
                    pages = {ascl:1710.017},
                    archivePrefix = {ascl},
                    eprint = {1710.017},
                    adsurl = {https://ui.adsabs.harvard.edu/abs/2017ascl.soft10017K},
                    adsnote = {Provided by the SAO/NASA Astrophysics Data System}
                }
                @ARTICLE{2002A&A...392..619H,
                    author = {{Heiter}, U. and {Kupka}, F. and {van't Veer-Menneret}, C. and {Barban}, C. and {Weiss}, W.~W. and {Goupil}, M. -J. and {Schmidt}, W. and {Katz}, D. and {Garrido}, R.},
                    title = "{New grids of ATLAS9 atmospheres I: Influence of convection treatments on model structure and on observable quantities}",
                    journal = {\aap},
                    keywords = {stars: atmospheres, stars: fundamental parameters, stars: variables delta Scuti stars, convection, Astrophysics},
                    year = 2002,
                    month = sep,
                    volume = {392},
                    pages = {619-636},
                    doi = {10.1051/0004-6361:20020788},
                    archivePrefix = {arXiv},
                    eprint = {astro-ph/0206156},
                    primaryClass = {astro-ph},
                    adsurl = {https://ui.adsabs.harvard.edu/abs/2002A&A...392..619H},
                    adsnote = {Provided by the SAO/NASA Astrophysics Data System}
                }
            """
        elif "ll" in self.source:
            self.citation_info = r"""
                @ARTICLE{2004A&A...428..993S,
                    author = {{Shulyak}, D. and {Tsymbal}, V. and {Ryabchikova}, T. and {St{\"u}tz}, Ch. and {Weiss}, W.~W.},
                    title = "{Line-by-line opacity stellar model atmospheres}",
                    journal = {\aap},
                    keywords = {stars: atmospheres, stars: abundances, stars: chemically peculiar, stars: fundamental parameters, stars: individual: CU Vir, stars: individual: HD 124224},
                    year = 2004,
                    month = dec,
                    volume = {428},
                    pages = {993-1000},
                    doi = {10.1051/0004-6361:20034169},
                    adsurl = {https://ui.adsabs.harvard.edu/abs/2004A&A...428..993S},
                    adsnote = {Provided by the SAO/NASA Astrophysics Data System}
                }
            """
        else:
            self.citation_info = ""  # ???

        atmo_grid = data["atmo_grid"]

        if "RADIUS" in atmo_grid.dtype.names and "HEIGHT" in atmo_grid.dtype.names:
            self.geom = "SPH"
            self["radius"] = atmo_grid["radius"]
            self["height"] = np.stack(atmo_grid["height"])
            # If the radius is given in absolute values
            # self["radius"] /= np.max(self["radius"])
        else:
            self.geom = "PP"
            self["radius"][:] = 1

        self.abund_format = "sme"

        # Scalar Parameters (one per atmosphere)
        self["teff"] = atmo_grid["teff"]
        self["logg"] = atmo_grid["logg"]
        self["monh"] = atmo_grid["monh"]
        self["vturb"] = atmo_grid["vturb"]
        self["lonh"] = atmo_grid["lonh"]
        self["wlstd"] = atmo_grid["wlstd"]
        # Vector Parameters (one array per atmosphere)
        self["rhox"] = np.stack(atmo_grid["rhox"])
        self["tau"] = np.stack(atmo_grid["tau"])
        self["temp"] = np.stack(atmo_grid["temp"])
        self["rho"] = np.stack(atmo_grid["rho"])
        self["xne"] = np.stack(atmo_grid["xne"])
        self["xna"] = np.stack(atmo_grid["xna"])
        self["abund"] = np.stack(atmo_grid["abund"])
        self["opflag"] = np.stack(atmo_grid["opflag"])

        # Store in cache
        cls._cache = self
        # And also replace the IDL file with a numpy file in the cache
        # We have to use a try except block, as this will crash with
        # permissions denied on windows, when trying to copy an open file
        # here the temporary file
        # Therefore we close the file, after copying and then delete it manually
        if lfs is not None:
            try:
                with NamedTemporaryFile(delete=False) as named:
                    self.save(named)
                    named.flush()
                lfs.move_to_cache(named.name, key=lfs.get_url(self.source))
            finally:
                try:
                    os.remove(named.name)
                except:
                    pass

        return self



# -*- coding: utf-8 -*-
"""
Contains functions to read and modify echelle structures, just as in reduce

Mostly for compatibility reasons
"""

import logging

import astropy.io.fits as fits
import numpy as np
import scipy.constants

logger = logging.getLogger(__name__)


def calc_2dpolynomial(solution2d):
    """Expand a 2d polynomial, where the data is given in a REDUCE make_wave format
    Note that the coefficients are for order/100 and column/1000 respectively, where the order is counted from order base up

    Parameters
    ----------
    solution2d : array
        data in a REDUCE make_wave format:
        0: version
        1: number of columns
        2: number of orders
        3: order base, i.e. 0th order number
        4-6: empty
        7: number of cross coefficients
        8: number of column only coefficients
        9: number of order only coefficients
        10: coefficient - constant
        11-x: column coefficients
        x-y : order coefficients
        z-  : cross coefficients (xy, xy2, x2y, x2y2, xy3, x3y), with x = orders, y = columns

    Returns
    -------
    poly : array[nord, ncol]
        expanded polynomial
    """

    # make wave style 2d fit
    ncol = int(solution2d[1])
    nord = int(solution2d[2])
    order_base = int(solution2d[3])
    deg_cross, deg_column, deg_order = (
        int(solution2d[7]),
        int(solution2d[8]),
        int(solution2d[9]),
    )
    coeff_in = solution2d[10:]

    coeff = np.zeros((deg_order + 1, deg_column + 1))
    coeff[0, 0] = coeff_in[0]
    coeff[0, 1:] = coeff_in[1 : 1 + deg_column]
    coeff[1:, 0] = coeff_in[1 + deg_column : 1 + deg_column + deg_order]
    coeff[1, 1] = coeff_in[deg_column + deg_order + 1]
    coeff[1, 2] = coeff_in[deg_column + deg_order + 2]
    coeff[2, 1] = coeff_in[deg_column + deg_order + 3]
    coeff[2, 2] = coeff_in[deg_column + deg_order + 4]

    if deg_cross == 6:
        coeff[1, 3] = coeff_in[deg_column + deg_order + 5]
        coeff[3, 1] = coeff_in[deg_column + deg_order + 6]

    x = np.arange(order_base, order_base + nord, dtype=float)
    y = np.arange(ncol, dtype=float)

    poly = np.polynomial.polynomial.polygrid2d(x / 100, y / 1000, coeff) / x[:, None]

    return poly


def calc_1dpolynomials(ncol, poly):
    """Expand a set of 1d polynomials (one per order) seperately

    Parameters
    ----------
    ncol : int
        number of columns
    poly : array[nord, degree]
        polynomial coefficients

    Returns
    -------
    poly : array[nord, ncol]
        expanded polynomials
    """

    nord = poly.shape[0]
    x = np.arange(ncol)
    result = np.zeros((nord, ncol))
    for i, coef in enumerate(poly):
        result[i] = np.polyval(coef, x)
    return result


def expand_polynomial(ncol, poly):
    """Checks if and how to expand data poly, then expands the data if necessary

    Parameters
    ----------
    ncol : int
        number of columns in the image
    poly : array[nord, ...]
        polynomial coefficients to expand, or already expanded data

    Returns
    -------
    poly : array[nord, ncol]
        expanded data
    """

    if poly.ndim == 1:
        poly = calc_2dpolynomial(poly)
    elif poly.shape[1] < 20:
        poly = calc_1dpolynomials(ncol, poly)
    return poly


def read(
    fname,
    extension=1,
    raw=False,
    continuum_normalization=True,
    barycentric_correction=True,
    radial_velociy_correction=True,
):
    """
    Read data from an echelle file
    Expand wavelength and continuum polynomials
    Apply barycentric/radial velocity correction
    Apply continuum normalization

    Will load any fields in the binary table, however special attention is given only to specific names:
    "SPEC"    : Spectrum
    "SIG"     : Sigma, i.e. (absolute) uncertainty
    "CONT"    : Continuum
    "WAVE"    : Wavelength solution
    "COLUMNS" : Column range

    Parameters
    ----------
    fname : str
        filename to load
    extension : int, optional
        fits extension of the data within the file (default: 1)
    raw : bool, optional
        if true apply no corrections to the data (default: False)
    continuum_normalization : bool, optional
        apply continuum normalization (default: True)
    barycentric_correction : bool, optional
        apply barycentric correction (default: True)
    radial_velociy_correction : bool, optional
        apply radial velocity correction (default: True)

    Returns
    -------
    ech : obj
        Echelle structure, with data contained in attributes
    """

    hdu = fits.open(fname)
    header = hdu[0].header
    data = hdu[extension].data

    ech = lambda: None  # placeholder
    ech.filename = fname
    ech.head = header

    for column in data.dtype.names:
        setattr(ech, column.lower(), data[column][0])
        if column == "SPEC":
            nord, ncol = data[column][0].shape

    if not raw:
        if hasattr(ech, "spec"):
            ech.orders = np.arange(nord)

        # Wavelength
        if hasattr(ech, "wave"):
            ech.wave = expand_polynomial(ncol, ech.wave)

            # Correct for radial velocity and barycentric correction
            velocity_correction = 0
            if barycentric_correction:
                velocity_correction -= header.get("barycorr", 0)
                header["barycorr"] = 0
            if radial_velociy_correction:
                velocity_correction += header.get("radvel", 0)
                header["radvel"] = 0

            speed_of_light = scipy.constants.speed_of_light * 1e-3
            ech.wave *= 1 + velocity_correction / speed_of_light

        if hasattr(ech, "cont"):
            ech.cont = expand_polynomial(ncol, ech.cont)

        # Create Mask, based on column range
        if hasattr(ech, "columns"):
            ech.mask = np.full((nord, ncol), False)
            for iord in range(nord):
                ech.mask[iord, : ech.columns[iord, 0]] = True
                ech.mask[iord, ech.columns[iord, 1] :] = True

            if hasattr(ech, "spec"):
                ech.spec = np.ma.masked_array(ech.spec, mask=ech.mask)
            if hasattr(ech, "sig"):
                ech.sig = np.ma.masked_array(ech.sig, mask=ech.mask)
            if hasattr(ech, "cont"):
                ech.cont = np.ma.masked_array(ech.cont, mask=ech.mask)
            if hasattr(ech, "wave"):
                ech.wave = np.ma.masked_array(ech.wave, mask=ech.mask)

        # Apply continuum normalization
        if hasattr(ech, "cont") and continuum_normalization:
            if hasattr(ech, "spec"):
                ech.spec /= ech.cont
            if hasattr(ech, "sig"):
                ech.sig /= ech.cont

    return ech


def save(fname, header, **kwargs):
    """Save data in an Echelle fits, i.e. a fits file with a Binary Table in Extension 1

    The data is passed in kwargs, with the name of the binary table column as the key
    Floating point data is saved as float32 (E), Integer data as int16 (I)

    Parameters
    ----------
    fname : str
        filename
    header : fits.header
        FITS header
    **kwargs : array[]
        data to be saved in the file
    """

    primary = fits.PrimaryHDU(header=header)

    columns = []
    for key, value in kwargs.items():
        arr = value.flatten()[None, :]

        if issubclass(arr.dtype.type, np.floating):
            arr = arr.astype(np.float32)
            dtype = "E"
        elif issubclass(arr.dtype.type, np.integer):
            arr = arr.astype(np.int16)
            dtype = "I"

        if np.ma.is_masked(arr):
            arr = np.ma.getdata(arr)

        form = "%i%s" % (value.size, dtype)
        dim = str(value.shape[::-1])
        columns += [fits.Column(name=key.upper(), array=arr, format=form, dim=dim)]

    table = fits.BinTableHDU.from_columns(columns)

    hdulist = fits.HDUList(hdus=[primary, table])
    hdulist.writeto(fname, overwrite=True)



# -*- coding: utf-8 -*-
"""
Elemental abundance data handling module
"""
import io
import json
import logging

import numpy as np
from flex.extensions.bindata import BinaryDataExtension

from .persistence import IPersist

logger = logging.getLogger(__name__)

_citation_asplund2009 = r"""
@ARTICLE{2009ARA&A..47..481A,
    author = {{Asplund}, Martin and {Grevesse}, Nicolas and {Sauval}, A. Jacques and
    {Scott}, Pat},
    title = "{The Chemical Composition of the Sun}",
    journal = {\araa},
    keywords = {Astrophysics - Solar and Stellar Astrophysics, Astrophysics - Earth and Planetary Astrophysics},
    year = "2009",
    month = "Sep",
    volume = {47},
    number = {1},
    pages = {481-522},
    doi = {10.1146/annurev.astro.46.060407.145222},
    archivePrefix = {arXiv},
    eprint = {0909.0948},
    primaryClass = {astro-ph.SR},
    adsurl = {https://ui.adsabs.harvard.edu/abs/2009ARA&A..47..481A},
    adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
"""

_citation_grevesse2007 = r"""
@ARTICLE{2007SSRv..130..105G,
    author = {{Grevesse}, N. and {Asplund}, M. and {Sauval}, A.~J.},
    title = "{The Solar Chemical Composition}",
    journal = {\ssr},
    keywords = {Sun: abundances, photosphere, corona},
    year = "2007",
    month = "Jun",
    volume = {130},
    number = {1-4},
    pages = {105-114},
    doi = {10.1007/s11214-007-9173-7},
    adsurl = {https://ui.adsabs.harvard.edu/abs/2007SSRv..130..105G},
    adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
"""

_citation_lodders2003 = r"""
@ARTICLE{2003ApJ...591.1220L,
    author = {{Lodders}, Katharina},
    title = "{Solar System Abundances and Condensation Temperatures of the Elements}",
    journal = {\apj},
    keywords = {Astrochemistry, Meteors, Meteoroids, Solar System: Formation- Sun: Abundances, Sun: Photosphere},
    year = "2003",
    month = "Jul",
    volume = {591},
    number = {2},
    pages = {1220-1247},
    doi = {10.1086/375492},
    adsurl = {https://ui.adsabs.harvard.edu/abs/2003ApJ...591.1220L},
    adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}
"""

_citation_atomic_weights = r"""
@article{loss2003atomic,
  title={Atomic weights of the elements 2001 (IUPAC Technical Report)},
  author={Loss, RD},
  journal={Pure and Applied Chemistry},
  volume={75},
  number={8},
  pages={1107--1122},
  year={2003},
  publisher={De Gruyter}
}
"""

# fmt: off
elements = (
    "H", "He",
    "Li", "Be", "B" , "C" , "N" , "O" , "F" , "Ne",
    "Na", "Mg", "Al", "Si", "P" , "S" , "Cl", "Ar",
    "K" , "Ca", "Sc", "Ti", "V" , "Cr", "Mn", "Fe",
    "Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se",
    "Br", "Kr", "Rb", "Sr", "Y" , "Zr", "Nb", "Mo",
    "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn",
    "Sb", "Te", "I" , "Xe", "Cs", "Ba", "La", "Ce",
    "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy",
    "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W" ,
    "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb",
    "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th",
    "Pa", "U" , "Np", "Pu", "Am", "Cm", "Bk", "Cf",
    "Es",)

# index of each element in the pattern data array
elements_dict = {el:i for i, el in enumerate(elements)}

# Array of atomic weights
# From Loss, R. D. 2003, Pure Appl. Chem., 75, 1107, "Atomic Weights
# of the Elements 2001". The isotopic composition of the Sun and stars
# may differ from the terrestrial values listed here!
# For radionuclides that appear only in Table 3, we have adopted the
# atomic mass of the longest lived isotope, rounding to at most two
# digits after the decimal point. These elements are: 43_Tc_98,
# 61_Pm_145, 85_At_210, 86_Rn_222, 87_Fr_223, 88_Ra_226, 89_Ac_227,
# 93_Np_237, 94_Pu_244, 95_Am_243, 96_Cm_247, 97_Bk_247, 98_Cf_251,
# 99_Es_252

_atom_weight = (
    1.00794, 4.002602,
    6.941, 9.012182, 10.811, 12.0107, 14.0067, 15.9994, 18.9984032, 20.1797,
    22.989770, 24.3050, 26.981538, 28.0855, 30.973761, 32.065, 35.453, 39.948,
    39.0983, 40.078, 44.955910, 47.867, 50.9415, 51.9961, 54.938049, 55.845,
    58.933200, 58.6934, 63.546, 65.409, 69.723, 72.64, 74.92160, 78.96,
    79.904, 83.798, 85.4678, 87.62, 88.90585, 91.224, 92.90638,
    95.94, 97.91, 95.94, 101.07, 102.90550, 106.42, 107.8682, 112.411,
    114.818, 118.710, 121.760, 127.60, 126.90447, 131.293, 132.90545, 137.327,
    138.9055, 140.116, 140.90765, 144.24, 144.91, 150.36, 151.964, 157.25,
    158.92534, 162.500, 164.93032, 167.259, 168.93421, 173.04, 174.967, 178.49,
    180.9479, 183.84, 186.207, 190.23, 192.217, 195.078, 196.966, 200.59,
    204.3833, 207.2, 208.98038, 209.99, 222.02, 223.02, 226.03, 227.03,
    232.0381, 231.03588, 238.02891, 237.05, 244.06, 243.06, 247.07, 247.07,
    251.08, 252.08,)

# Asplund, Grevesse, Sauval, Scott (2009,  Annual Review of Astronomy
# and Astrophysics, 47, 481)
_asplund2009 = (
    12.00, 10.93,
    1.05, 1.38, 2.70, 8.43, 7.83, 8.69, 4.56,  7.93,
    6.24,  7.60,  6.45,  7.51,  5.41,  7.12,  5.50,  6.40,
    5.03,  6.34,  3.15,  4.95,  3.93,  5.64,  5.43,  7.50,
    4.99,  6.22,  4.19,  4.56,  3.04,  3.65,  2.30,  3.34,
    2.54,  3.25,  2.52,  2.87,  2.21,  2.58,  1.46,  1.88,
    None,  1.75,  0.91,  1.57,  0.94,  1.71,  0.80,  2.04,
    1.01,  2.18,  1.55,  2.24,  1.08,  2.18,  1.10,  1.58,
    0.72,  1.42,  None,  0.96,  0.52,  1.07,  0.30,  1.10,
    0.48,  0.92,  0.10,  0.84,  0.10,  0.85,  -0.12,  0.85,
    0.26,  1.40,  1.38,  1.62,  0.92,  1.17,  0.90,  1.75,
    0.65,  None,  None,  None,  None,  None,  None,  0.02,
    None,  -0.54,  None,  None,  None,  None,  None,  None,
    None,)

# Grevesse, Asplund, Sauval (2007, Space Science Review, 130, 105)
_grevesse2007 = (
    12.00, 10.93,
    1.05, 1.38, 2.70, 8.39, 7.78, 8.66, 4.56, 7.84,
    6.17, 7.53, 6.37, 7.51, 5.36, 7.14, 5.50, 6.18,
    5.08, 6.31, 3.17, 4.90, 4.00, 5.64, 5.39, 7.45,
    4.92, 6.23, 4.21, 4.60, 2.88, 3.58, 2.29, 3.33,
    2.56, 3.25, 2.60, 2.92, 2.21, 2.58, 1.42, 1.92,
    None, 1.84, 1.12, 1.66, 0.94, 1.77, 1.60, 2.00,
    1.00, 2.19, 1.51, 2.24, 1.07, 2.17, 1.13, 1.70,
    0.58, 1.45, None, 1.00, 0.52, 1.11, 0.28, 1.14,
    0.51, 0.93, 0.00, 1.08, 0.06, 0.88, -0.17, 1.11,
    0.23, 1.25, 1.38, 1.64, 1.01, 1.13, 0.90, 2.00,
    0.65, None, None, None, None, None, None, 0.06,
    None, -0.52, None, None, None, None, None, None,
    None,)

# Lodders 2003 (ApJ, 591, 1220)
_lodders2003 = (
    12, 10.899,
    3.28, 1.41, 2.78, 8.39, 7.83, 8.69, 4.46, 7.87,
    6.30, 7.55, 6.46, 7.54, 5.46, 7.19, 5.26, 6.55,
    5.11, 6.34, 3.07, 4.92, 4.00, 5.65, 5.50, 7.47,
    4.91, 6.22, 4.26, 4.63, 3.10, 3.62, 2.32, 3.36,
    2.59, 3.28, 2.36, 2.91, 2.20, 2.60, 1.42, 1.96,
    1.82, 1.11, 1.70, 1.23, 1.74, 0.80, 2.11, 1.06,
    2.22, 1.54, 2.27, 1.10, 2.18, 1.18, 1.61, 0.75,
    1.46, 0.95, 0.85, 1.06, 0.31, 1.13, 0.49, 0.95,
    0.11, 0.94, 0.09, 0.77, -0.14, 0.65, 0.26, 1.37,
    1.35, 1.67, 0.72, 1.16, 0.81, 2.05, 0.68, 0.09,
    None, -0.49, None, None, None, None, None, None,
    None, None, None, None, None, None, None, None,
    None,)
# fmt: on


class Abund(IPersist):
    """Elemental abundance data and methods.
    Valid abundance pattern types are:

    'sme' - For hydrogen, the abundance value is the fraction of all
        nuclei that are hydrogen, including all ionization states
        and treating molecules as constituent atoms. For the other
        elements, the abundance values are log10 of the fraction of
        nuclei of each element in any form relative to the total for
        all elements in any form. For the Sun, the abundance values
        of H, He, and Li are approximately 0.92, -1.11, and -11.0.
    'n/nTot' - Abundance values are the fraction of nuclei
        of each element in any form relative to the total for all
        elements in any form. For the Sun, the abundance values of
        H, He, and Li are approximately 0.92, 0.078, and 1.03e-11.
    'n/nH' - Abundance values are the fraction of nuclei
        of each element in any form relative to the number of
        hydrogen nuclei in any form. For the Sun, the abundance
        values of H, He, and Li are approximately 1, 0.085, and
        1.12e-11.
    'H=12' - Abundance values are log10 of the fraction of nuclei of
        each element in any form relative to the number of hydrogen
        in any form plus an offset of 12. For the Sun, the nuclei
        abundance values of H, He, and Li are approximately 12,
        10.9, and 1.05.
    """

    def __init__(
        self,
        monh=0,
        pattern="solar",
        type="sme",
        citation_info=_citation_atomic_weights,
    ):
        self.monh = monh
        # The internal type is fixed to this value
        self._type_internal = "H=12"
        self.type = type
        self.citation_info = citation_info
        if isinstance(pattern, str):
            self.set_pattern_by_name(pattern)
        else:
            self.set_pattern_by_value(pattern, type)

    def __call__(self, type="H=12", raw=False):
        """Return abundances for all elements.
        Apply current [M/H] value to the current abundance pattern.
        Transform the resulting abundances to the requested abundance type.
        """
        pattern = np.copy(self._pattern)
        if self.monh is not None:
            pattern[2:] += self.monh
        return self.totype(pattern, type, raw=raw, copy=False)

    def __getitem__(self, elem):
        return self.get_element(elem)

    def __setitem__(self, elem, abund):
        self.update_pattern({elem: abund})

    def __str__(self):
        if self._pattern is None:
            if self._monh is None:
                return "[M/H] is not set. Abundance pattern is not set."
            else:
                return f"[M/H]={self._monh:.3f}. Abundance pattern is not set"
        else:
            a = self.get_pattern("H=12", raw=True)
            if self._monh is None:
                out = "[M/H] is not set. Values below are the abundance pattern.\n"
            else:
                a = np.copy(a)
                a[2:] += self._monh
                out = (
                    f" [M/H]={self._monh:.3f} applied to abundance pattern. "
                    "Values below are abundances.\n"
                )
            for i in range(9):
                for j in range(11):
                    out += "  {:<5s}".format(elements[11 * i + j])
                out += "\n"
                for j in range(11):
                    out += "{:7.3f}".format(a[11 * i + j])
                if i < 8:
                    out += "\n"
            return out

    def __repr__(self):
        return self.__str__()

    # These are there for the atmosphere interpolation
    # The internal format is always H=12, so that should be fine
    def __add__(self, other):
        result = Abund(
            monh=self.monh, pattern=np.copy(self._pattern), type=self._type_internal
        )
        if isinstance(other, Abund):
            result._pattern += other.get_pattern(self._type_internal, raw=True)
            result.monh += other.monh
        else:
            raise NotImplementedError
        result.type = self.type
        return result

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        result = Abund(
            monh=self.monh, pattern=np.copy(self._pattern), type=self._type_internal
        )
        if isinstance(other, Abund):
            raise NotImplementedError
        else:
            result._pattern *= other
            result.monh *= other
        result.type = self.type
        return result

    def __rmul__(self, other):
        return self.__mul__(other)

    def _get_index(self, elem):
        try:
            return elements_dict[elem]
        except KeyError:
            raise KeyError(
                f"Got element abreviation {elem}, expected one of {elements}"
            )

    def __copy__(self):
        pattern = self._pattern.copy()
        other = Abund(
            monh=self.monh,
            type=self.type,
            pattern=pattern,
            citation_info=self.citation_info,
        )
        return other

    @staticmethod
    def fromtype(pattern, fromtype, raw=False):
        """Return a copy of the input abundance pattern, transformed from
        the input type to the 'H=12' type. Valid abundance pattern types
        are 'sme', 'n/nTot', 'n/nH', 'n/nFe', and 'H=12'.
        """
        elem = elements

        if isinstance(pattern, dict):
            abund = np.array([pattern[el] for el in elem], dtype=float)
        else:
            abund = np.array(pattern, dtype=float)

        if np.isnan(abund[0]):
            raise ValueError("Pattern must define abundance of H")

        type = fromtype.lower()
        if type == "h=12":
            pass
        elif type == "sme":
            abund[1:] += 12 - np.log10(abund[0])
            abund[0] = 12
        elif type == "n/ntot":
            abund /= abund[0]
            abund = 12 + np.log10(abund)
        elif type == "n/nh":
            abund = 12 + np.log10(abund)
        elif type == "n/nfe":
            abund /= abund[0]
            abund = 12 + np.log10(abund)
        elif type == "fe=12":
            abund = 10 ** (abund - 12)
            abund /= abund[0]
            abund = 12 + np.log10(abund)
        else:
            raise ValueError(
                "got abundance type '{}',".format(type)
                + " should be 'H=12', 'n/nH', 'n/nTot', 'n/nFe', 'Fe=12', or 'sme'"
            )
        if raw:
            return abund
        else:
            return {el: abund[elements_dict[el]] for el in elements}

    @staticmethod
    def totype(pattern, totype, raw=False, copy=True):
        """Return a copy of the input abundance pattern, transformed from
        the 'H=12' type to the output type. Valid abundance pattern types
        are 'sme', 'n/nTot', 'n/nH', and 'H=12'.
        """
        if isinstance(pattern, dict):
            abund = [pattern[el] if el in pattern.keys() else np.nan for el in elements]
            abund = np.array(abund)
        elif copy:
            abund = np.copy(pattern)
        else:
            abund = pattern

        type = totype.lower()

        if type == "h=12":
            pass
        elif type == "sme":
            abund2 = 10 ** (abund - 12)
            abund[0] = 1 / np.nansum(abund2)
            abund[1:] = abund[1:] - 12 + np.log10(abund[0])
            # abund /= np.sum(abund)
            # abund[1:] = np.log10(abund[1:])
        elif type == "n/ntot":
            abund = 10 ** (abund - 12)
            abund /= np.nansum(abund)
        elif type == "n/nh":
            abund = 10 ** (abund - 12)
        elif type == "n/nfe":
            idx_fe = elements_dict["Fe"]
            abund = 10 ** (abund - 12)
            abund /= abund[idx_fe]
        elif type == "fe=12":
            idx_fe = elements_dict["Fe"]
            abund = 10 ** (abund - 12)
            abund /= abund[idx_fe]
            abund = np.log10(abund) + 12
        else:
            raise ValueError(
                "got abundance type '{}',".format(type)
                + " should be 'H=12', 'n/nH', 'n/nTot', 'n/nFe', 'Fe=12', or 'sme'"
            )

        if raw:
            return abund
        else:
            return {el: abund[elements_dict[el]] for el in elements}

    _formats = ["H=12", "sme", "n/nTot", "n/nH", "n/nFe", "Fe=12"]

    @property
    def elem(self):
        """Return the standard abbreviation for each element.
        Use property so user will not redefine elements.
        """
        return elements

    @property
    def elem_dict(self):
        """Return the position of each element in the raw array"""
        return elements_dict

    @property
    def monh(self):
        """float: Metallicity, the logarithmic offset added to
        the abundance pattern for all elements except hydrogen and helium."""
        return self._monh

    @monh.setter
    def monh(self, monh):
        self._monh = float(monh)

    @property
    def pattern(self):
        """array: Abundance pattern in the initial format"""
        return self.get_pattern(self.type, raw=False)

    def set_pattern_by_name(self, pattern_name):
        """Set the abundance pattern to one of the predefined options

        Parameters
        ----------
        pattern_name : str
            Name of the predefined option to use. One of 'asplund2009', 'grevesse2007',
            'lodders2003', 'empty'

        Raises
        ------
        ValueError
            If an undefined pattern_name was given
        """
        self.type = "H=12"
        self.citation_info = _citation_atomic_weights + "\n"
        if pattern_name.lower() in ["asplund2009"]:
            self._pattern = np.array(_asplund2009, dtype=float)
            self.citation_info += _citation_asplund2009
        elif pattern_name.lower() in ["grevesse2007", "solar"]:
            self._pattern = np.array(_grevesse2007, dtype=float)
            self.citation_info += _citation_grevesse2007
        elif pattern_name.lower() == "lodders2003":
            self._pattern = np.array(_lodders2003, dtype=float)
            self.citation_info += _citation_lodders2003
        elif pattern_name.lower() == "empty":
            self._pattern = self.empty_pattern()
        else:
            raise ValueError(
                f"Got abundance pattern name {pattern_name} should be one of"
                "'asplund2009', 'grevesse2007', 'lodders2003', or 'empty'."
            )

    def set_pattern_by_value(self, pattern, type):
        self._pattern = self.fromtype(pattern, type, raw=True)
        self.type = type

    def update_pattern(self, updates):
        """Update the abundance pattern for several elements at once

        The abundance is first converted into the initially specified format,
        before being converted back to the internal format

        Parameters
        ----------
        updates : dict{str:float}
            the elements to update

        Raises
        ------
        KeyError
            If any of the element keys is not valid
        """
        pattern = self.get_pattern(type=self.type, raw=True)
        for key in updates:
            pos = self._get_index(key)
            pattern[pos] = updates[key]
        self.set_pattern_by_value(pattern, self.type)

    def get_pattern(self, type=None, raw=False):
        """
        Transform the specified abundance pattern from type used
        internally by SME to the requested type. Valid abundance pattern
        types are: 'sme', 'n/nTot', 'n/nH', 'H=12'

        Parameters
        ----------
        type : str
            The pattern format to retrieve the pattern as. Defaults to the
            original input format.
        raw : bool
            If True will return the pattern as a numpy array, with indices as defined in elem dict.
            Otherwise the return value is a dictionary, with the elements as keys.
        """
        if type is None:
            type = self.type
        return self.totype(self._pattern, type, raw=raw)

    def get_element(self, elem, type=None):
        if type is None:
            type = self.type
        pattern = self(type, raw=True)
        i = self._get_index(elem)
        return pattern[i]

    def empty_pattern(self):
        """Return an abundance pattern with value None for all elements."""
        pattern = np.full(len(elements), np.nan)
        pattern[0] = 0
        return pattern

    def to_dict(self):
        data = {
            "monh": self.monh,
            "type_internal": self._type_internal,
            "type": self.type,
            "citation_info": self.citation_info,
            "data": self._pattern,
        }
        return data

    @classmethod
    def from_dict(cls, data):
        obj = cls(
            monh=data["monh"],
            pattern=data["data"],
            type=data["type_internal"],
            citation_info=data["citation_info"],
        )
        obj.type = data["type"]
        return obj

    def _save(self):
        header = {
            "monh": self.monh,
            "type_internal": self._type_internal,
            "type": self.type,
            "citation_info": self.citation_info,
        }
        data = self._pattern
        ext = BinaryDataExtension(header, data)
        return ext

    @classmethod
    def _load(cls, ext: BinaryDataExtension):
        header = ext.header
        pattern = ext.data
        abund = cls(
            monh=header["monh"],
            pattern=pattern,
            type=header["type"],
            citation_info=header["citation_info"],
        )
        return abund

    def _save_v1(self, file, folder="abund"):
        """Save the data to a file handler"""
        if folder != "" or folder[-1] != "/":
            folder += "/"

        monh = float(self.monh) if self.monh is not None else None
        info = {
            "format": self.type,
            "monh": monh,
            "citation_info": self.citation_info,
        }
        file.writestr(f"{folder}info.json", json.dumps(info))

        b = io.BytesIO()
        np.save(b, self.get_pattern(raw=True))
        file.writestr(f"{folder}pattern.npy", b.getvalue())

    @staticmethod
    def _load_v1(file, names, folder=""):
        """Load the data from a file handler"""
        for name in names:
            if name.endswith("info.json"):
                info = file.read(name)
                info = json.loads(info)

                abund_format = info["format"]
                monh = info["monh"]
                citation_info = info.get("citation_info", _citation_atomic_weights)

            elif name.endswith("pattern.npy"):
                b = io.BytesIO(file.read(name))
                pattern = np.load(b)

        return Abund(
            monh=monh,
            pattern=pattern,
            type=abund_format,
            citation_info=citation_info,
        )

    @staticmethod
    def solar():
        """Return solar abundances of asplund 2009"""
        solar = Abund(pattern="solar", monh=0)
        return solar



# -*- coding: utf-8 -*-
import re
from os.path import basename

import numpy as np

from ..abund import Abund
from .atmosphere import Atmosphere


class KrzFile(Atmosphere):
    """Read .krz atmosphere files"""

    def __init__(self, filename, source=None):
        super().__init__()
        if source is None:
            self.source = basename(filename)
        else:
            self.source = source
        self.method = "embedded"
        self.citation_info = r"""
            @MISC{2017ascl.soft10017K,
                author = {{Kurucz}, Robert L.},
                title = "{ATLAS9: Model atmosphere program with opacity distribution functions}",
                keywords = {Software},
                year = "2017",
                month = "Oct",
                eid = {ascl:1710.017},
                pages = {ascl:1710.017},
                archivePrefix = {ascl},
                eprint = {1710.017},
                adsurl = {https://ui.adsabs.harvard.edu/abs/2017ascl.soft10017K},
                adsnote = {Provided by the SAO/NASA Astrophysics Data System}}
        """
        self.load(filename)

    def load(self, filename):
        """
        Load data from disk

        Parameters
        ----------
        filename : str
            name of the file to load
        """
        # TODO: this only works for some krz files
        # 1..2 lines header
        # 3 line opacity
        # 4..13 elemntal abundances
        # 14.. Table data for each layer
        #    Rhox Temp XNE XNA RHO

        with open(filename, "r") as file:
            header1 = file.readline()
            header2 = file.readline()
            opacity = file.readline()
            abund = [file.readline() for _ in range(10)]
            table = file.readlines()

        # Combine the first two lines
        header = header1 + header2
        # Parse header
        # vturb

        try:
            self.vturb = float(re.findall(r"VTURB=?\s*(\d)", header, flags=re.I)[0])
        except IndexError:
            self.vturb = 0

        try:
            self.lonh = float(re.findall(r"L/H=?\s*(\d+.?\d*)", header, flags=re.I)[0])
        except IndexError:
            self.lonh = 0

        self.teff = float(re.findall(r"T ?EFF=?\s*(\d+.?\d*)", header, flags=re.I)[0])
        self.logg = float(
            re.findall(r"GRAV(ITY)?=?\s*(\d+.?\d*)", header, flags=re.I)[0][1]
        )

        model_type = re.findall(r"MODEL TYPE=?\s*(\d)", header, flags=re.I)[0]
        self.model_type = int(model_type)

        model_type_key = {0: "rhox", 1: "tau", 3: "sph"}
        self.depth = model_type_key[self.model_type]
        self.geom = "pp"

        self.wlstd = float(re.findall(r"WLSTD=?\s*(\d+.?\d*)", header, flags=re.I)[0])
        # parse opacity
        i = opacity.find("-")
        opacity = opacity[:i].split()
        self.opflag = np.array([int(k) for k in opacity])

        # parse abundance
        pattern = np.genfromtxt(abund).flatten()[:-1]
        pattern[1] = 10 ** pattern[1]
        self.abund = Abund(monh=0, pattern=pattern, type="sme")

        # parse table
        self.table = np.genfromtxt(table, delimiter=",", usecols=(0, 1, 2, 3, 4))
        self.rhox = self.table[:, 0]
        self.temp = self.table[:, 1]
        self.xne = self.table[:, 2]
        self.xna = self.table[:, 3]
        self.rho = self.table[:, 4]



# -*- coding: utf-8 -*-
from distutils.core import Extension, setup
from os.path import abspath, dirname, join

import numpy.distutils.misc_util
from libtools import get_full_libfile

libdir = get_full_libfile()
libdir = abspath(dirname(libdir))
include_dirs = numpy.distutils.misc_util.get_numpy_include_dirs()
include_dirs += [libdir]

module = Extension(
    "_smelib",
    sources=["_smelib.cpp"],
    language="c++",
    include_dirs=include_dirs,
    libraries=["sme"],
    library_dirs=[libdir],
)

setup(ext_modules=[module])



# -*- coding: utf-8 -*-
import logging

import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage.filters import convolve

logger = logging.getLogger(__name__)


def apply_broadening(ipres, x_seg, y_seg, type="gauss", sme=None):
    """
    Broaden a spectrum by instrument resolution, with a given broadening type

    Parameters
    ----------
    ipres : float
        instrument resolution
    x_seg : array (npoints,)
        x values (wavelength) of the spectrum to broaden
    y_seg : array (npoints,)
        y values (intensities) of the spectrum to broaden
    type : {str, None}, optional
        broadening type to apply. Options are "gauss", "sinc", "table", None.
        If None, will try to use sme.iptype to determine type. "table" requires
        keyword sme to be passed as well. See functions the respective for details.
        (default: "gauss")
    sme : SME_Struct, optional
        sme structure with instrument profile data, required only for
        type="table" or type=None (default: None)

    Raises
    ------
    AttributeError
        if type requires SME_Struct, but its missing
        passed type not recognized

    Returns
    -------
    y_seg : array (npoints,)
        broadened intensity spectrum
    """

    if sme is None and type in ["table", None]:
        raise AttributeError(f"SME structure needs to be passed when using type={type}")

    # Using the log-linear wavelength grid requires using the first point
    # for specifying the width of the instrumental profile
    hwhm = 0.5 * x_seg[0] / ipres if ipres > 0 else 0

    if type is None:
        type = sme.iptype
    type = type.casefold()

    if type == "table":
        y_seg = tablebroad(x_seg, y_seg, sme.ip_x, sme.ip_y)
    elif type == "gauss":
        y_seg = gaussbroad(x_seg, y_seg, hwhm)
    elif type == "sinc":
        y_seg = sincbroad(x_seg, y_seg, hwhm)
    else:
        raise AttributeError(f"Unknown instrument profile type - {type}")

    return y_seg


def tablebroad(w, s, xip, yip):
    """
    Convolves a spectrum with an arbitrary instrumental profile.

    Parameters
    -------
    w : array of size (n,)
        wavelength scale of spectrum to be smoothed
    s : array of size (n,)
        spectrum to be smoothed
    xip : array of size (m,)
        x points of the instrument profile
    yip : array of size (m,)
        y points of the instrument profile
    Returns
    -------
    sout: array[n]
        the smoothed spectrum.
    """
    """
    History
    -------
        22-May-92 JAV
            Switched instrumental profile from multiple gaussians
            to gaussian with power-law wings.
        04-Aug-92 JAV
            Renamed from ipsmo.pro# changed f/ procedure to function.
            Switched f/ 10 to 15 Hamilton pixels in each wing.
        20-Oct-92 JAV
            Switched from gpfunc to ipfun (4 to 5 par).
        23-Aug-94 JAV
            Switched to explicitly passed IPs.
        Oct-18  AW
            Python version
    """

    # Define sizes
    dsdh = np.abs(np.min(np.diff(xip)))
    nip = 2 * int(15 / dsdh) + 1  ## profile points

    # Generate instrumental profile on model pixel scale.
    x = (
        np.arange(nip, dtype=float) - (nip - 1) / 2
    ) * dsdh  # offset in Hamilton pixels
    ip = interp1d(xip, yip, kind="cubic")(x)
    # ip = bezier_interp(xip, yip, x)  # spline onto new scale
    ip = ip[::-1]  # reverse for convolution
    ip = ip / np.sum(ip)  # ensure unit area

    # Pad spectrum ends to minimize impact of Fourier ringing.
    sout = convolve(s, ip, mode="nearest")

    return sout  # return convolved spectrum


def gaussbroad(w, s, hwhm):
    """
    Smooths a spectrum by convolution with a gaussian of specified hwhm.

    Parameters
    -------
    w : array[n]
        wavelength scale of spectrum to be smoothed
    s : array[n]
        spectrum to be smoothed
    hwhm : float
        half width at half maximum of smoothing gaussian.

    Returns
    -------
    sout: array[n]
        the gaussian-smoothed spectrum.
    """
    """
    History
    --------
        Dec-90 GB,GM
            Rewrote with fourier convolution algorithm.
        Jul-91 AL
            Translated from ANA to IDL.
        22-Sep-91 JAV
            Relaxed constant dispersion check# vectorized, 50% faster.
        05-Jul-92 JAV
            Converted to function, handle nonpositive hwhm.
        Oct-18 AW
            Python version
    """

    # Warn user if hwhm is negative.
    if hwhm < 0:
        logger.warning("Forcing negative smoothing width to zero.")

    # Return input argument if half-width is nonpositive.
    if hwhm <= 0:
        return s  # true: no broadening

    # Calculate (uniform) dispersion.
    nw = len(w)  ## points in spectrum
    wrange = w[-1] - w[0]
    dw = wrange / (nw - 1)  # wavelength change per pixel

    # Make smoothing gaussian; extend to 4 sigma.
    # 4.0 / sqrt(2.0 * alog(2.0)) = 3.3972872
    # sqrt(alog(2.0)) = 0.83255461
    # sqrt(alog(2.0) / pi) = 0.46971864
    # (*1.0000632 to correct for >4 sigma wings)
    if hwhm >= 5 * wrange:
        return np.full(nw, np.sum(s) / nw)
    ## points in half gaussian
    nhalf = int(3.3972872 * hwhm / dw)
    ## points in gaussian (odd!)
    ng = 2 * nhalf + 1
    # wavelength scale of gaussian
    wg = dw * (np.arange(ng, dtype=float) - (ng - 1) / 2)
    # convenient absisca
    xg = (0.83255461 / hwhm) * wg
    # unit area gaussian w / FWHM
    gpro = (0.46974832 * dw / hwhm) * np.exp(-xg * xg)
    gpro = gpro / np.sum(gpro)

    # Pad spectrum ends to minimize impact of Fourier ringing.
    sout = convolve(s, gpro, mode="nearest")

    return sout


def sincbroad(w, s, hwhm):
    """
    Smooths a spectrum by convolution with a sinc function of specified hwhm.

    Parameters
    ------
    w : array of size (n,)
        wavelength scale of spectrum to be smoothed
    s : array of size (n,)
        spectrum to be smoothed
    hwhm : float
        half width at half maximum of smoothing gaussian.

    Returns
    -------
    sout : array of size (n,)
        the sinc-smoothed spectrum.
    """
    """
    History
    -------
    Dec-90 GB,GM
        Rewrote with fourier convolution algorithm.
    Jul-91 AL
        Translated from ANA to IDL.
    22-Sep-91 JAV
        Relaxed constant dispersion check# vectorized, 50% faster.
    05-Jul-92 JAV
        Converted to function, handle nonpositive hwhm.
    14-Nov-93 JAV
        Adapted from macbro.pro
    23-Apr-93 JAV
        Verified that convolution kernel has specified hwhm. For IR FTS
        spectra: hwhm=0.0759 Angstroms, max change in profile is 0.4% of continuum.
    Oct-18 AW
        Python Version
    """

    # Warn user if hwhm is negative.
    if hwhm < 0:
        logger.warning("Forcing negative smoothing width to zero.")

    # Return input argument if half-width is nonpositive.
    if hwhm <= 0:
        return s  # true: no broadening

    # Calculate (uniform) dispersion.
    nw = len(w)  ## points in spectrum
    dw = (w[-1] - w[0]) / (nw - 1)  # wavelength change per pixel

    # Make sinc function out to 20th zero-crossing on either side. Error due to
    # ignoring additional lobes is less than 0.2% of continuum. Reducing extent
    # to 10th zero-crossing doubles maximum error.
    fwhm = 2.0 * hwhm  # full width at half maximum
    rperfw = 0.26525  # radians per fwhm of sinc
    xrange = 20 * np.pi  # 20th zero of sinc (radians)
    wrange = xrange * fwhm * rperfw  # 20th zero of sinc (wavelength)
    nhalf = int(wrange / dw + 0.999)  ## points in half sinc
    nsinc = 2 * nhalf + 1  ## points in sinc (odd!)
    wsinc = (np.arange(nsinc, dtype=float) - nhalf) * dw  # absissca (wavelength)
    xsinc = wsinc / (fwhm * rperfw)  # absissca (radians)
    xsinc[nhalf] = 1.0  # avoid divide by zero
    sinc = np.sin(xsinc) / xsinc  # calculate sinc
    sinc[nhalf] = 1.0  # insert midpoint
    xsinc[nhalf] = 0.0  # fix xsinc
    sinc = sinc / np.sum(sinc)  # normalize sinc

    # Pad spectrum ends to minimize impact of Fourier ringing.
    sout = convolve(s, sinc, mode="nearest")

    return sout



# -*- coding: utf-8 -*-
"""
System to store large data files on a server
Load them whem required by the user
Update the pointer file on github when new datafiles become available

Pro: Versioning is effectively done by Git
Con: Need to run server
"""

import gzip
import json
import logging
import os
from os.path import basename
from pathlib import Path
from tempfile import NamedTemporaryFile

from astropy.utils.data import (
    clear_download_cache,
    download_file,
    import_file_to_cache,
    is_url_in_cache,
)
from tqdm.auto import tqdm
from tqdm.utils import CallbackIOWrapper

from .config import Config
from .util import show_progress_bars

logger = logging.getLogger(__name__)

# We are lazy and want a simple check if a file is in the Path
Path.__contains__ = lambda self, key: (self / key).exists()

PKGNAME = "sme"


class LargeFileStorage:
    """
    Download large data files from data server when needed
    New versions of the datafiles are indicated in a 'pointer' file
    that includes the hash of the newest version of the files

    Raises
    ------
    FileNotFoundError
        If the datafiles can't be located anywhere
    """

    def __init__(self, server, pointers, storage):
        #:Server: Large File Storage Server address
        self.server = server

        if isinstance(pointers, str):
            path = Path(__file__).parent / pointers
            pointers = LargeFileStorage.load_pointers_file(path)

        #:dict(fname:hash): points from a filename to the current newest object id, usually a hash
        self.pointers = pointers
        #:Directory: directory of the current data files
        self.current = Path(storage).expanduser().absolute()

    @staticmethod
    def load_pointers_file(filename):
        try:
            with open(str(filename), "r") as f:
                pointers = json.load(f)
        except FileNotFoundError:
            logger.error("Could not find LargeFileStorage reference file %s", filename)
            pointers = {}
        return pointers

    def get(self, key):
        """
        Request a datafile from the LargeFileStorage
        Assures that tracked files are at the specified version
        And downloads data from the server if necessary

        Parameters
        ----------
        key : str
            Name of the requested datafile

        Raises
        ------
        FileNotFoundError
            If the requested datafile can not be found anywhere

        Returns
        -------
        fullpath : str
            Absolute path to the datafile
        """
        url = self.get_url(key)
        # If its a direct file link, pass that directly to
        if url.startswith("file://"):
            return url[7:]

        fname = download_file(url, cache=True, pkgname=PKGNAME)

        compression = self._test_compression(fname)
        if compression is None:
            pass
        elif compression == "gzip":
            # If the file is compressed
            # Replace the cache file with the decompressed file
            self._unpack_gzip(fname, key, url)
        else:
            raise ValueError(
                "The file is compressed using %s, which is not supported" % compression
            )

        return fname

    def _test_compression(self, fname):
        """Check filetype using the magic string"""
        with open(fname, "rb") as f:
            magic = f.read(6)
            if magic[:2] == b"\x1f\x8b":
                return "gzip"
            if magic[:6] == b"\x37\x7A\xBC\xAF\x27\x1C":
                return "7z"
            if magic[:5] == b"\x50\x4B\x03\x04":
                return "zip"
            return None

    def _unpack_gzip(self, fname, key, url):
        logger.debug("Unpacking data file %s", key)

        # We have to use a try except block, as this will crash with
        # permissions denied on windows, when trying to copy an open file
        # here the temporary file
        # Therefore we close the file, after copying and then delete it manually
        try:
            with gzip.open(fname, "rb") as f_in:
                with NamedTemporaryFile("wb", delete=False) as f_out:
                    with tqdm(
                        # total=f_in.size,
                        desc="Unpack",
                        unit="B",
                        unit_scale=True,
                        unit_divisor=1024,
                        disable=~show_progress_bars,
                    ) as t:
                        fobj = CallbackIOWrapper(t.update, f_in, "read")
                        while True:
                            chunk = fobj.read(1024)
                            if not chunk:
                                break
                            f_out.write(chunk)
                        f_out.flush()
                        t.reset()
            import_file_to_cache(url, f_out.name, pkgname=PKGNAME)
        finally:
            try:
                os.remove(f_out.name)
            except:
                pass

    def get_url(self, key):
        key = str(key)

        # Check if the file is tracked and/or exists in the storage directory
        if key not in self.pointers:
            if key not in self.current:
                if not os.path.exists(key):
                    raise FileNotFoundError(
                        f"File {key} does not exist and is not tracked by the Large File system"
                    )
                else:
                    return Path(key).as_uri()
            else:
                return (self.current / key).as_uri()

        # Otherwise get it from the cache or online if necessary
        newest = self.pointers[key]
        url = self.server + "/" + newest
        return url

    def clean_cache(self):
        """Remove unused cache files (from old versions)"""
        clear_download_cache(pkgname=PKGNAME)

    def delete_file(self, fname):
        """Delete a file, including the cache file"""
        clear_download_cache(fname, pkgname=PKGNAME)

    def move_to_cache(self, fname, key=None):
        """Move currently used files into cache directory and use symlinks instead,
        just as if downloaded from a server"""
        if key is None:
            key = basename(fname)
        import_file_to_cache(key, fname, pkgname=PKGNAME)
        self.pointers[key] = key


def setup_atmo(config=None):
    if config is None:
        config = Config()
    server = config["data.file_server"]
    storage = config["data.atmospheres"]
    pointers = config["data.pointers.atmospheres"]
    lfs_atmo = LargeFileStorage(server, pointers, storage)
    return lfs_atmo


def setup_nlte(config=None):
    if config is None:
        config = Config()
    server = config["data.file_server"]
    storage = config["data.nlte_grids"]
    pointers = config["data.pointers.nlte_grids"]
    lfs_nlte = LargeFileStorage(server, pointers, storage)
    return lfs_nlte


def setup_lfs(config=None, lfs_atmo=None, lfs_nlte=None):
    if config is None:
        config = Config()
    if lfs_atmo is None:
        lfs_atmo = setup_atmo(config)
    if lfs_nlte is None:
        lfs_nlte = setup_nlte(config)
    return config, lfs_atmo, lfs_nlte


def get_available_atmospheres(config=None):
    if config is None:
        config = Config()
    pointers = config["data.pointers.atmospheres"]
    storage = config["data.atmospheres"]
    data = get_available_files(pointers, storage)
    return data


def get_available_nlte_grids(config=None):
    if config is None:
        config = Config()
    pointers = config["data.pointers.nlte_grids"]
    storage = config["data.nlte_grids"]
    data = get_available_files(pointers, storage)
    return data


def get_available_files(pointers, storage):
    pointers = Path(__file__).parent / pointers
    storage = Path(storage).expanduser()
    data = LargeFileStorage.load_pointers_file(pointers)
    files = list(data.keys())
    files_non_lfs = [
        f
        for f in os.listdir(storage)
        if f not in data and not os.path.isdir(storage / f)
    ]
    files += files_non_lfs
    return files
